#!/usr/bin/env python
# ! -*- coding: utf-8 -*-


import os
import sys
module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path)

import numpy as np
import pandas as pd
from autobmt.feature_binning import bin_method_run, chi_bin, dt_bin, equal_freq_bin, kmeans_bin
#from .feature_binning import bin_method_run, chi_bin, dt_bin, equal_freq_bin, kmeans_bin

np.random.seed(1)
feature = np.random.rand(500)
t = np.random.randint(2, size=500)
Aa = np.random.randint(100, size=500)
Bb = np.random.randint(3, size = 500)

df = pd.DataFrame({
    'feature': feature,
    'target': t,
    'A': Aa,
})

LENGTH = 500
np.random.seed(1)
A = np.random.rand(LENGTH)
A[np.random.choice(LENGTH, 20, replace=False)] = np.nan

B = np.random.randint(100, size=LENGTH)
C = A + np.random.normal(0, 0.2, LENGTH)
D = A + np.random.normal(0, 0.1, LENGTH)

E = np.random.rand(LENGTH)
E[np.random.choice(LENGTH, 480, replace=False)] = np.nan

F = B + np.random.normal(0, 10, LENGTH)

target = np.random.randint(2, size=LENGTH)

frame = pd.DataFrame({
    'A': A,
    'B': B,
    'C': C,
    'D': D,
    'E': E,
    'F': F,
})

frame['target'] = target


def test_chimerge():
    splits = chi_bin(feature, t, n_bins=10, is_need_monotonic=False)
    print(splits)
    assert len(splits) == 9


def test_chimerge_bins_not_enough():
    splits = chi_bin(Bb, t, n_bins=10, is_need_monotonic=False)
    assert len(splits) == 2


def test_chimerge_bins_with_min_samples():
    splits = chi_bin(feature, t, n_bins=15, min_sample_rate=0.02, is_need_monotonic=False)
    print(len(splits))
    assert len(splits) == 10


def test_dtmerge():
    splits = dt_bin(feature, t, n_bins=10, is_need_monotonic=False)
    assert len(splits) == 9


def test_quantilemerge():
    splits = equal_freq_bin(feature, n_bins=10, is_need_monotonic=False)
    assert len(splits) == 9


def test_quantilemerge_not_enough():
    splits = equal_freq_bin(Bb, n_bins=10, is_need_monotonic=False)
    assert len(splits) == 2


# def test_stepmerge():
#     splits = StepMerge(feature, n_bins = 10)
#     assert len(splits) == 9

def test_kmeansmerge():
    splits = kmeans_bin(feature, n_bins=10, is_need_monotonic=False)
    assert len(splits) == 9


def test_merge():
    splits, res = bin_method_run(feature, target=t, method='chi', return_bin=True, n_bins=10,
                                 is_need_monotonic=False)
    assert len(np.unique(res)) == 10


def test_merge_frame():
    res = bin_method_run(df, target='target', method='chi', return_bin=True, n_bins=10, is_need_monotonic=False)
    assert len(np.unique(res['feature'])) == 10


def test_merge_frame2():
    res = bin_method_run(df, target='target', method='chi', n_bins=10, is_need_monotonic=False)
    assert len(np.unique(res['feature'])) == 10






print("======================测试equal_freq_bin")
def test_equal_freq_bin1():
    s = equal_freq_bin(D, target, n_bins=21, is_need_monotonic=False)
    print("=======len(s)======", len(s))
    assert list(s)[:-1] == list(
        [0.012419821725897555, 0.08160251414710648, 0.1365937204837973, 0.20126781529042354, 0.26078161584175596,
         0.3043923889835296, 0.37010564491684994, 0.43393496112637275, 0.4761375755670674, 0.5346413336617297,
         0.5827083140162775, 0.623464641428682, 0.6750798738210625, 0.7170075229286542, 0.77298483315698,
         0.8173532653547075, 0.8517076655269138, 0.9129958438409241, 0.9850361692249885])

def test_equal_freq_bin2():
    s = equal_freq_bin(D, target, n_bins=10, is_need_monotonic=False)
    assert len(s) == 10

def test_equal_freq_bin3():
    s = equal_freq_bin(D, target, n_bins=21)
    assert list(s)[:-1] == list([0.5346413336617297])

def test_equal_freq_bin4():
    ##测默认箱子
    s = equal_freq_bin(D, target, is_need_monotonic=False)
    assert len(s) == 9

def test_equal_freq_bin5():
    s = bin_method_run(frame[['D', 'target']], target='target', method='equal_freq', n_bins=21, is_need_monotonic=True)
    assert len(np.unique(s['D'])) == 3

def test_equal_freq_bin6():
    s = bin_method_run(frame[['D', 'target']], target='target', method='equal_freq', n_bins=21)
    assert len(np.unique(s['D'])) == 3

def test_equal_freq_bin7():
    s = bin_method_run(frame[['D', 'target']], target='target', method='equal_freq', n_bins=21, is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 21

def test_equal_freq_bin8():
    s = bin_method_run(frame[['D', 'target']], target='target', method='equal_freq')
    assert len(np.unique(s['D'])) == 3

def test_equal_freq_bin9():
    s = bin_method_run(frame[['D', 'target']], target='target', method='equal_freq', n_bins=21)
    assert len(np.unique(s['D'])) == 3

def test_equal_freq_bin10():
    s = bin_method_run(frame[['D', 'target']], target, method='equal_freq', n_bins=21, is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 21

def test_equal_freq_bin11():
    s = bin_method_run(frame[['D', 'target']], target=None, method='equal_freq', n_bins=21, is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 21

def test_equal_freq_bin12():
    s = bin_method_run(frame[['D', 'target']], target=None, method='equal_freq', n_bins=21)
    assert len(np.unique(s['D'])) == 21

def test_equal_freq_bin13():
    ##测默认箱子
    s = bin_method_run(frame[['D', 'target']], target, method='equal_freq', is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 10

def test_equal_freq_bin14():
    s = bin_method_run(frame[['D', 'target']].fillna(-9), target, method='equal_freq', is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 10

print("======================测试equal_freq_bin")
print("======================测试kmeans")
def test_kmeans_bin1():
    s = kmeans_bin(D, target, n_bins=21, is_need_monotonic=False)
    print("=======len(s)======", len(s))
    assert list(s)[:-1] == list(
        [0.05394007952037447, 0.1886799394187435, 0.31125728976809286, 0.4206303104301733, 0.5315341201997051,
         0.6423777917886719, 0.7568052721344689, 0.8707566397289682, 0.9905607587443476])

def test_kmeans_bin2():
    s = kmeans_bin(D, target, n_bins=10, is_need_monotonic=False)
    print("=======len(s)======", len(s))
    assert len(s) == 10

def test_kmeans_bin3():
    s = kmeans_bin(D, target, n_bins=21)
    assert list(s)[:-1] == list([0.491896259821787])

def test_kmeans_bin4():
    ##测默认箱子
    s = kmeans_bin(D, target, is_need_monotonic=False)
    assert len(s) == 9

def test_kmeans_bin5():
    s = bin_method_run(frame[['D', 'target']], target='target', method='kmeans', n_bins=21, is_need_monotonic=True)
    assert len(np.unique(s['D'])) == 3

def test_kmeans_bin6():
    s = bin_method_run(frame[['D', 'target']], target='target', method='kmeans', n_bins=21)
    assert len(np.unique(s['D'])) == 3

def test_kmeans_bin7():
    s = bin_method_run(frame[['D', 'target']], target='target', method='kmeans', n_bins=21, is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 11

def test_kmeans_bin8():
    s = bin_method_run(frame[['D', 'target']], target='target', method='kmeans')
    assert len(np.unique(s['D'])) == 3

def test_kmeans_bin9():
    s = bin_method_run(frame[['D', 'target']], target='target', method='kmeans', n_bins=21)
    assert len(np.unique(s['D'])) == 3

def test_kmeans_bin10():
    s = bin_method_run(frame[['D', 'target']], target, method='kmeans', n_bins=21, is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 11

def test_kmeans_bin11():
    s = bin_method_run(frame[['D', 'target']], target=None, method='kmeans', n_bins=21, is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 11

def test_kmeans_bin12():
    s = bin_method_run(frame[['D', 'target']], target=None, method='kmeans', n_bins=21)
    assert len(np.unique(s['D'])) == 11

def test_kmeans_bin13():
    ##测默认箱子
    s = bin_method_run(frame[['D', 'target']], target, method='kmeans', is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 10

def test_kmeans_bin14():
    s = bin_method_run(frame[['B', 'target']], target, method='kmeans', is_need_monotonic=False)
    assert len(np.unique(s['B'])) == 10
print("======================测试kmeans")
print("======================测试dt")

def test_dt_bin1():
    s = dt_bin(D, target, n_bins=21, is_need_monotonic=False)
    print("=======len(s)======", len(s))
    assert list(s)[:-1] == list(
        [0.03355235606431961, 0.14810966700315475, 0.2149350941181183, 0.3048485219478607, 0.37758709490299225,
         0.44229236245155334, 0.49767398834228516, 0.5686621367931366, 0.6086502075195312, 0.6941415071487427,
         0.7509821057319641, 0.8093653619289398, 0.843919038772583, 0.899420440196991, 0.9890741109848022])

def test_dt_bin2():
    s = dt_bin(D, target, n_bins=10, is_need_monotonic=False)
    assert len(s) == 10

def test_dt_bin3():
    s = dt_bin(D, target, n_bins=21)
    assert list(s)[:-1] == list([0.9890741109848022])

def test_dt_bin4():
    ##测默认箱子
    s = dt_bin(D, target, is_need_monotonic=False)
    assert len(s) == 9

def test_dt_bin5():
    s = bin_method_run(frame[['D', 'target']], target='target', method='dt', n_bins=21, is_need_monotonic=True)
    assert len(np.unique(s['D'])) == 3

def test_dt_bin6():
    s = bin_method_run(frame[['D', 'target']], target='target', method='dt', n_bins=21)
    assert len(np.unique(s['D'])) == 3

def test_dt_bin7():
    s = bin_method_run(frame[['D', 'target']], target='target', method='dt', n_bins=21, is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 17

def test_dt_bin8():
    s = bin_method_run(frame[['D', 'target']], target='target', method='dt')
    assert len(np.unique(s['D'])) == 3

def test_dt_bin9():
    s = bin_method_run(frame[['D', 'target']], target, method='dt', n_bins=21, is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 17

def test_dt_bin10():
    ##测默认箱子
    s = bin_method_run(frame[['D', 'target']], target, method='dt', is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 10

def test_dt_bin11():
    s = bin_method_run(frame[['D', 'target']].fillna(-9), target, method='dt', is_need_monotonic=False)
    assert len(np.unique(s['D'])) == 10
print("======================测试dt")
print("======================测试chi")
def test_chi_bin1():
    s = chi_bin(A, target, n_bins=21, is_need_monotonic=False)
    print("=======len(s)======", len(s))
    assert list(s)[:-1] == list([0.1354279030721699, 0.4108113499221856, 0.7438258540750929, 0.8763891522960383])

def test_chi_bin2():
    s = chi_bin(A, target, n_bins=4, is_need_monotonic=False)
    assert len(s) == 4

def test_chi_bin3():
    s = chi_bin(A, target, n_bins=13, is_need_monotonic=False)
    assert len(s) == 5

def test_chi_bin4():
    s = chi_bin(A, target, n_bins=21)
    print("=======len(s)======", len(s))
    assert list(s)[:-1] == list([0.7438258540750929])

def test_chi_bin5():
    ##测默认箱子
    s = chi_bin(A, target, is_need_monotonic=False)
    print("=======len(s)======", len(s))
    assert len(s) == 5

def test_chi_bin6():
    s = bin_method_run(frame[['A', 'target']], target='target', method='chi', n_bins=21, is_need_monotonic=True)
    assert len(np.unique(s['A'])) == 3

def test_chi_bin7():
    s = bin_method_run(frame[['A', 'target']], target='target', method='chi', n_bins=21)
    assert len(np.unique(s['A'])) == 3

def test_chi_bin8():
    s = bin_method_run(frame[['A', 'target']], target='target', method='chi', n_bins=21, is_need_monotonic=False)
    assert len(np.unique(s['A'])) == 6

def test_chi_bin9():
    s = bin_method_run(frame[['A', 'target']], target='target', method='chi')
    assert len(np.unique(s['A'])) == 3

def test_chi_bin10():
    s = bin_method_run(frame[['A', 'target']], target, method='chi', n_bins=21, is_need_monotonic=False)
    assert len(np.unique(s['A'])) == 6

def test_chi_bin11():
    ##测默认箱子
    s = bin_method_run(frame[['A', 'target']], target, method='chi', is_need_monotonic=False)
    assert len(np.unique(s['A'])) == 6

def test_chi_bin12():
    s = bin_method_run(frame[['A', 'target']].fillna(-9), target, method='chi', is_need_monotonic=False)
    assert len(np.unique(s['A'])) == 5
print("======================测试chi")




# #####测试决策树分箱
# def split_empty(feature, y=None, return_mask=True):
#     copy_fea = np.copy(feature)
#     mask = pd.isna(copy_fea)
#
#     copy_fea = copy_fea[~mask]
#     if y is not None:
#         copy_y = np.copy(y)
#         copy_y = copy_y[~mask]
#         return copy_fea, copy_y, mask
#     return copy_fea, mask
#
# from sklearn.tree import DecisionTreeClassifier, _tree
# tree = DecisionTreeClassifier(
#     min_samples_leaf=1,
#     max_leaf_nodes=22,
#     # 优先满足min_samples_leaf参数。在满足min_samples_leaf参数参数后，再考虑max_leaf_nodes。
#     # 比如情况1：min_samples_leaf设置成0.05，max_leaf_nodes设置成20。满足0.05后，最大max_leaf_nodes只有10，那也就这样了
#     # 比如情况2：min_samples_leaf设置成0.05，max_leaf_nodes设置成6。满足0.05后，最大max_leaf_nodes有10，那再考虑max_leaf_nodes，继续分到满足max_leaf_nodes=6停止
#     # ps:min_samples_leaf=1表示没有限制
# )
#
# feature, target, empty_mask = split_empty(D, target)
#
# tree.fit(feature.reshape((-1, 1)), target)
# thresholds = tree.tree_.threshold
# thresholds = thresholds[thresholds != _tree.TREE_UNDEFINED]
# splits = np.sort(thresholds)
# print(len(splits))
# #####测试决策树分箱



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: report2excel.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-05-20
'''

import re
import string
import warnings

import xlsxwriter

from .logger_utils import Logger

warnings.filterwarnings('ignore')

log = Logger(level="info", name=__name__).logger


class Report2Excel:
    """
    统计信息写入到excel
    """

    def __init__(self, file_name, workbook=None, current_row=8, row_spaces=3, column_space=2):
        self.file_name = file_name
        self.workbook = self.create_workbook(workbook)
        self.define_global_format()
        self.row_spaces = row_spaces
        self.column_space = column_space

    def create_workbook(self, workbook):
        """类初始化的时候,必须创建或者返回一个workbook对象"""
        if isinstance(workbook, xlsxwriter.workbook.Workbook):
            return workbook
        else:
            workbook = xlsxwriter.Workbook(self.file_name, {'nan_inf_to_errors': True})
        return workbook

    def close_workbook(self):
        """关闭workbook对象,开始写入到excel文件中"""
        self.workbook.close()

    def df_to_excel(self, evaluate_df, sheet_name):
        # excel相关格式
        worksheet = self.workbook.add_worksheet(sheet_name)

        def write_df_to_excel_by_row(df, worksheet, start_row, start_column, str_formats):
            """按列写入df,因为要单独给每一列一种格式"""
            for col_num, column_name in enumerate(df.columns.values):
                try:
                    worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                           str_formats)
                except Exception as e:
                    t = list(map(str, df[column_name].values.tolist()))
                    worksheet.write_column(start_row, start_column + col_num, t,
                                           str_formats)

        # 写入标题
        for col_num, value in enumerate(evaluate_df.columns.values):
            worksheet.write(0, col_num, value, self.header_format)

        # 写入数据和插入图像
        start_row, start_column = 1, 0
        write_df_to_excel_by_row(evaluate_df, worksheet, start_row, start_column, self.str_formats)
        worksheet.set_column(0, 1, 18)
        # workbook.close()

    def define_global_format(self):
        """定义Excel全局格式"""
        self.merge_format = self.workbook.add_format({  # 大标题的格式
            'bold': True,
            'font_name': '微软雅黑',
            'font_size': 20,
            'border': False,  # 边框线
            'align': 'center',  # 水平居中
            'valign': 'vcenter',  # 垂直居中
            'fg_color': '#ffffff',  # 颜色填充
        })
        self.title_format = self.workbook.add_format({  # 标题的格式
            'bold': True,
            'font_name': '微软雅黑',
            'font_size': 15,
            'border': False,  # 边框线
            'valign': 'top',
            # 'fg_color': '#ddebf7'
        })
        self.sub_title_format = self.workbook.add_format({  # 子标题的格式
            'bold': True,
            'font_name': '微软雅黑',
            'font_size': 12,
            'border': False,  # 边框线
            'valign': 'top',
            # 'fg_color': '#ddebf7'
        })
        self.content_format = self.workbook.add_format({  # 文字的格式
            'bold': False,
            'font_name': '微软雅黑',
            'font_size': 10,
            'border': False,  # 边框线
            'valign': 'top',
        })
        self.content_dict_format = self.workbook.add_format({  # 文字的格式
            'bold': True,
            'font_name': '微软雅黑',
            'font_size': 10,
            'border': False,  # 边框线
            'valign': 'top',
            'font_color': '#275b8e',
        })
        self.ps_format = self.workbook.add_format({  # 注释的格式
            'bold': False,
            'font_name': '微软雅黑',
            'font_size': 10,
            'font_color': 'red',
            'border': False,  # 边框线
            'valign': 'top'
        })
        self.header_format = self.workbook.add_format({
            'bold': True,
            'text_wrap': True,
            'valign': 'center',
            'font_name': 'Consolas',
            'font_size': 11,
            'border': 1,
            'bottom': 2,
        })
        ### 数字格式
        self.left_formats = self.workbook.add_format({'font_name': 'Consolas', 'align': 'left'})
        self.integer_formats = self.workbook.add_format(
            {'font_name': 'Consolas', 'num_format': '#,##0', 'align': 'left'})  # 整数型
        self.float_formats = self.workbook.add_format(
            {'font_name': 'Consolas', 'num_format': '#,##0.00', 'align': 'left'})  # 浮点型保留2位
        self.percentage_formats = self.workbook.add_format(
            {'font_name': 'Consolas', 'num_format': '0.00%', 'align': 'left'})  # 百分数保留2位
        self.varname_formats = self.workbook.add_format(
            {'color': 'blue', 'underline': True, 'font_name': 'Consolas', 'font_size': 11, 'align': 'left'})
        self.str_formats = self.workbook.add_format(
            {'font_name': 'Consolas', 'font_size': 11, 'align': 'left'})
        self.float_formats4 = self.workbook.add_format(
            {'num_format': '#,##0.0000', 'font_name': 'Consolas', 'font_size': 11, 'align': 'left'})  # 浮点型保留4位
        self.font_formats = self.workbook.add_format({'font_name': 'Consolas', 'font_size': 11, 'align': 'left'})  # 其他
        self.float_formats_new = self.workbook.add_format(
            {'font_name': 'Consolas', 'num_format': '_(* #,##0.00_);_(* (#,##0.00);_(* "-"??_);_(@_)',
             'align': 'left'})  # 百分数保留2位


# Standard Library


# Third Party Stuff


EXCEL_RANGE_PATTERN = re.compile(r'([a-zA-Z]+)([\d]+):([a-zA-Z]+)([\d]+)')

XLSXWRITER_FORMAT_PROPERTIES = (
    'font_name',
    'font_size',
    'font_color',
    'bold',
    'italic',
    'underline',
    'font_strikeout',
    'font_script',
    'num_format',
    'locked',
    'hidden',
    'text_h_align',
    'text_v_align',
    'rotation',
    'text_wrap',
    'text_justlast',
    # 'center_across',
    'indent',
    'shrink',
    'pattern',
    'bg_color',
    'fg_color',
    'bottom',
    'top',
    'left',
    'right',
    'bottom_color',
    'top_color',
    'left_color',
    'right_color',
)


def duplicate_xlsxwriter_format_object(workbook, old_format):
    properties = {}
    if old_format is not None:
        for property_name in XLSXWRITER_FORMAT_PROPERTIES:
            properties[property_name] = getattr(old_format, property_name)

    return workbook.add_format(properties)


def col2num(col):
    num = 0
    for c in col:
        if c in string.ascii_letters:
            num = num * 26 + (ord(c.upper()) - ord('A')) + 1
    return num


def excel_range_string_to_indices(range_string):
    try:
        first_col_name, first_row, last_col_name, last_row = EXCEL_RANGE_PATTERN.findall(
            range_string)[0]
    except IndexError:
        raise ValueError("Invalid range string.")

    first_col_index = col2num(first_col_name) - 1
    first_row_index = int(first_row) - 1
    last_col_index = col2num(last_col_name) - 1
    last_row_index = int(last_row) - 1

    return (
        first_col_index,
        first_row_index,
        last_col_index,
        last_row_index
    )


def apply_border_to_cell(workbook, worksheet, row_index, col_index, format_properties):
    try:
        cell = worksheet.table[row_index][col_index]
        new_format = duplicate_xlsxwriter_format_object(workbook, cell.format)

        # Convert properties in the constructor to method calls.
        for key, value in format_properties.items():
            getattr(new_format, 'set_' + key)(value)

        # Update cell object
        worksheet.table[row_index][col_index] = cell = cell._replace(format=new_format)
    except KeyError:
        format = workbook.add_format(format_properties)
        worksheet.write(row_index, col_index, None, format)


def apply_outer_border_to_range(workbook, worksheet, options=None):
    options = options or {}

    border_style = options.get("border_style", 1)
    range_string = options.get("range_string", None)

    if range_string is not None:
        first_col_index, first_row_index, last_col_index, last_row_index = excel_range_string_to_indices(
            range_string)
    else:
        first_col_index = options.get("first_col_index", None)
        last_col_index = options.get("last_col_index", None)
        first_row_index = options.get("first_row_index", None)
        last_row_index = options.get("last_row_index", None)

        all_are_none = all(map(lambda x: x is None, [
            first_col_index,
            last_col_index,
            first_row_index,
            last_row_index,
        ]))

        if all_are_none:
            raise Exception("You need to specify the range")

    for row_index in range(first_row_index, last_row_index + 1):
        left_border = {
            "left": border_style,
        }
        right_border = {
            "right": border_style,
        }

        apply_border_to_cell(workbook, worksheet, row_index, first_col_index, left_border)
        apply_border_to_cell(workbook, worksheet, row_index, last_col_index, right_border)

    for col_index in range(first_col_index, last_col_index + 1):
        top_border = {
            "top": border_style,
        }

        bottom_border = {
            "bottom": border_style,
        }

        apply_border_to_cell(workbook, worksheet, first_row_index, col_index, top_border)
        apply_border_to_cell(workbook, worksheet, last_row_index, col_index, bottom_border)

    top_left_border = {
        "top": border_style,
        "left": border_style,
    }
    apply_border_to_cell(workbook, worksheet, first_row_index, first_col_index, top_left_border)

    top_right_border = {
        "top": border_style,
        "right": border_style,
    }
    apply_border_to_cell(workbook, worksheet, first_row_index, last_col_index, top_right_border)

    bottom_left_border = {
        "bottom": border_style,
        "left": border_style,
    }
    apply_border_to_cell(workbook, worksheet, last_row_index, first_col_index, bottom_left_border)

    bottom_right_border = {
        "bottom": border_style,
        "right": border_style,
    }
    apply_border_to_cell(workbook, worksheet, last_row_index, last_col_index, bottom_right_border)


def var_summary_to_excelold(var_summary_df, workbook=None, sheet_name=None):
    # excel相关格式
    # workbook = xlsxwriter.Workbook(output, {'nan_inf_to_errors': True})
    # worksheet = workbook.add_worksheet('分箱详情')
    worksheet = workbook.add_worksheet(sheet_name)
    # Add a header format.
    header_format = workbook.add_format({
        'bold': True,
        'text_wrap': True,
        'valign': 'center',
        'font_name': 'Consolas',
        'font_size': 11,
        'border': 1,
        'bottom': 2,
    })
    varname_formats = workbook.add_format(
        {'color': 'blue', 'underline': True, 'font_name': 'Consolas', 'font_size': 11, })
    percentage_formats = workbook.add_format(
        {'num_format': '0.00%', 'font_name': 'Consolas', 'font_size': 11, })  # 百分数保留2位
    pct_bad_formats = workbook.add_format(
        {'num_format': '0.00%', 'font_name': 'Consolas', 'font_size': 11, 'font_color': '#980101'})  # 百分数保留2位
    pct_good_formats = workbook.add_format(
        {'num_format': '0.00%', 'font_name': 'Consolas', 'font_size': 11, 'font_color': '#0050aa'})  # 百分数保留2位
    float_formats4 = workbook.add_format(
        {'num_format': '#,##0.0000', 'font_name': 'Consolas', 'font_size': 11, })  # 浮点型保留4位
    integer_formats = workbook.add_format(
        {'num_format': '#,##0', 'font_name': 'Consolas', 'font_size': 11, })  # 整数型
    font_formats = workbook.add_format({'font_name': 'Consolas', 'font_size': 11, })  # 整数型
    row_formats = workbook.add_format({'bottom': 1})  # 下边框线

    def colnum_string(n):
        """将数字n转成excel的列符号"""
        string = ""
        while n > 0:
            n, remainder = divmod(n - 1, 26)
            string = chr(65 + remainder) + string
        return string

    def write_df_to_excel_by_row(index, feature, df, workbook, worksheet, start_row, start_column):
        """按列写入df,因为要单独给每一列一种格式"""
        for col_num, column_name in enumerate(df.columns.values):
            if column_name in ['positive_rate', 'Pct_Bin']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       percentage_formats)
            elif column_name in ['Pct_Bad']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       pct_bad_formats)
            elif column_name in ['Pct_Good']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       pct_good_formats)
            elif column_name in ['var_name']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       varname_formats)
            elif column_name in ['iv', 'IV']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       float_formats4)
            elif column_name in ['Total', 'Bad', 'Good']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       integer_formats)
            elif column_name in ['woe']:
                # 输入到Excel中的时候，将WoE的值放大100倍，方便观察
                # worksheet.write_column(start_row, start_column + col_num,
                #                        # (df[column_name] * 100).astype(int).values.tolist(),
                #                        integer_formats)
                worksheet.write_column(start_row, start_column + col_num,
                                       df[column_name].values.tolist(),
                                       float_formats4)
            else:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       font_formats)

        # 给最后一行加上下框线,用于区分变量
        end_row, end_column = start_row + df.shape[0], df.shape[1]
        worksheet.conditional_format('A{end_row}:{end_column}{end_row}' \
                                     .format(end_row=end_row, end_column=colnum_string(end_column)), \
                                     {'type': 'no_blanks', 'format': row_formats})

        # 给BadRate列加上条件格式
        badrate_index = colnum_string(start_column + df.columns.get_loc('Bad_Rate') + 1)
        worksheet.conditional_format(
            '{c}{start_row}:{c}{end_row}'.format(start_row=start_row, end_row=end_row, c=badrate_index),
            {'type': 'data_bar', 'bar_color': '#f0c4c4', 'bar_solid': True})

        # 给每个分箱变量,绘制badrate曲线和woe分布图
        # badrate
        graph = workbook.add_chart({'type': 'line'})
        range_column_index = df.columns.get_loc('range') + start_column
        badrate_column_index = df.columns.get_loc('Bad_Rate') + start_column

        range_categories = [worksheet.get_name(), start_row, range_column_index, end_row - 1,
                            range_column_index]
        range_value = [worksheet.get_name(), start_row, badrate_column_index, end_row - 1,
                       badrate_column_index]
        graph.add_series({'name': 'BadRate', 'categories': range_categories, 'values': range_value,
                          'marker': {'type': 'circle'}, 'data_labels': {'value': True}})
        graph.set_size({'width': 450, 'height': 200})
        graph.set_title(
            {'name': 'Bad Rate - {}'.format(feature), 'overlay': False,
             'name_font': {'name': '微软雅黑', 'size': 9, 'bold': True}})
        graph.set_x_axis({'line': {'none': True}})
        graph.set_y_axis({'line': {'none': True}})
        graph.set_legend({'none': True})  # 设置图例
        worksheet.insert_chart(start_row + index * 7, end_column + 1, graph)

        # WoE
        column_graph = workbook.add_chart({'type': 'column'})
        range_column_index = df.columns.get_loc('range') + start_column
        badrate_column_index = df.columns.get_loc('woe') + start_column

        range_categories = [worksheet.get_name(), start_row, range_column_index, end_row - 1,
                            range_column_index]
        range_value = [worksheet.get_name(), start_row, badrate_column_index, end_row - 1,
                       badrate_column_index]
        column_graph.add_series({'name': 'woe', 'categories': range_categories, 'values': range_value,
                                 'marker': {'type': 'circle'}, 'data_labels': {'value': True}})
        column_graph.set_size({'width': 450, 'height': 200})  # TODO 大小改成和单元格的大小一致,以单元格的大小为单位
        column_graph.set_title(
            {'name': 'WoE - {}'.format(feature), 'overlay': False,
             'name_font': {'name': '微软雅黑', 'size': 9, 'bold': True}})
        column_graph.set_x_axis({'line': {'none': True}, 'label_position': 'low'})
        column_graph.set_y_axis({'line': {'none': True}})
        column_graph.set_legend({'none': True})  # 设置图例
        worksheet.insert_chart(start_row + index * 7, end_column + 9, column_graph)

        # 给变量名称插入跳转超链接
        for i in range(1, df.shape[0] + 1):
            worksheet.write_url('B{}'.format(start_row + i),
                                'internal:{}!M{}:AC{}'.format(worksheet.get_name(), start_row + index * 7 + 1,
                                                              start_row + index * 7 + 10), string=feature,
                                tip='Jump to charts')

    # 写入标题
    for col_num, value in enumerate(var_summary_df.columns.values):
        worksheet.write(0, col_num, value, header_format)

    # 写入数据和插入图像
    start_row, start_column = 1, 0
    index = 0
    for name, single_df in var_summary_df.groupby('var_name'):
        write_df_to_excel_by_row(index, name, single_df, workbook, worksheet, start_row, start_column)
        start_row = single_df.shape[0] + start_row
        index += 1

    # 冻结窗格
    worksheet.freeze_panes(1, 2)
    worksheet.set_column(0, 1, 18)

    # # 保存为excel
    # workbook.close()


def var_summary_to_excel(var_summary_df, workbook=None, sheet_name=None):
    # excel相关格式
    # workbook = xlsxwriter.Workbook(output, {'nan_inf_to_errors': True})
    # worksheet = workbook.add_worksheet('分箱详情')
    worksheet = workbook.add_worksheet(sheet_name)
    # Add a header format.
    header_format = workbook.add_format({
        'bold': True,
        'text_wrap': True,
        'valign': 'center',
        'font_name': 'Consolas',
        'font_size': 11,
        'border': 1,
        'bottom': 2,
    })
    varname_formats = workbook.add_format(
        {'color': 'blue', 'underline': True, 'font_name': 'Consolas', 'font_size': 11, })
    percentage_formats = workbook.add_format(
        {'num_format': '0.00%', 'font_name': 'Consolas', 'font_size': 11, })  # 百分数保留2位
    positive_pct_formats = workbook.add_format(
        {'num_format': '0.00%', 'font_name': 'Consolas', 'font_size': 11, 'font_color': '#980101'})  # 百分数保留2位
    negative_pct_formats = workbook.add_format(
        {'num_format': '0.00%', 'font_name': 'Consolas', 'font_size': 11, 'font_color': '#0050aa'})  # 百分数保留2位
    float_formats4 = workbook.add_format(
        {'num_format': '#,##0.0000', 'font_name': 'Consolas', 'font_size': 11, })  # 浮点型保留4位
    integer_formats = workbook.add_format(
        {'num_format': '#,##0', 'font_name': 'Consolas', 'font_size': 11, })  # 整数型
    font_formats = workbook.add_format({'font_name': 'Consolas', 'font_size': 11, })  # 整数型
    row_formats = workbook.add_format({'bottom': 1})  # 下边框线

    def colnum_string(n):
        """将数字n转成excel的列符号"""
        string = ""
        while n > 0:
            n, remainder = divmod(n - 1, 26)
            string = chr(65 + remainder) + string
        return string

    def write_df_to_excel_by_row(index, feature, df, workbook, worksheet, start_row, start_column):
        """按列写入df,因为要单独给每一列一种格式"""
        for col_num, column_name in enumerate(df.columns.values):
            if column_name in ['positive_rate', 'total_pct']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       percentage_formats)
            elif column_name in ['positive_pct']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       positive_pct_formats)
            elif column_name in ['negative_pct']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       negative_pct_formats)
            elif column_name in ['var_name']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       varname_formats)
            elif column_name in ['iv', 'IV']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       float_formats4)
            elif column_name in ['total', 'positive', 'negative']:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       integer_formats)
            elif column_name in ['woe']:
                # 输入到Excel中的时候，将WoE的值放大100倍，方便观察
                # worksheet.write_column(start_row, start_column + col_num,
                #                        # (df[column_name] * 100).astype(int).values.tolist(),
                #                        integer_formats)
                worksheet.write_column(start_row, start_column + col_num,
                                       df[column_name].values.tolist(),
                                       float_formats4)
            else:
                worksheet.write_column(start_row, start_column + col_num, df[column_name].values.tolist(),
                                       font_formats)

        # 给最后一行加上下框线,用于区分变量
        end_row, end_column = start_row + df.shape[0], df.shape[1]
        worksheet.conditional_format('A{end_row}:{end_column}{end_row}' \
                                     .format(end_row=end_row, end_column=colnum_string(end_column)), \
                                     {'type': 'no_blanks', 'format': row_formats})

        # 给BadRate列加上条件格式
        badrate_index = colnum_string(start_column + df.columns.get_loc('positive_rate') + 1)
        worksheet.conditional_format(
            '{c}{start_row}:{c}{end_row}'.format(start_row=start_row, end_row=end_row, c=badrate_index),
            {'type': 'data_bar', 'bar_color': '#f0c4c4', 'bar_solid': True})

        # 给每个分箱变量,绘制badrate曲线和woe分布图
        # badrate
        graph = workbook.add_chart({'type': 'line'})
        range_column_index = df.columns.get_loc('range') + start_column
        badrate_column_index = df.columns.get_loc('positive_rate') + start_column

        range_categories = [worksheet.get_name(), start_row, range_column_index, end_row - 1,
                            range_column_index]
        range_value = [worksheet.get_name(), start_row, badrate_column_index, end_row - 1,
                       badrate_column_index]
        graph.add_series({'name': 'positive_rate', 'categories': range_categories, 'values': range_value,
                          'marker': {'type': 'circle'}, 'data_labels': {'value': True}})
        graph.set_size({'width': 450, 'height': 200})
        graph.set_title(
            {'name': 'positive_rate - {}'.format(feature), 'overlay': False,
             'name_font': {'name': '微软雅黑', 'size': 9, 'bold': True}})
        graph.set_x_axis({'line': {'none': True}})
        graph.set_y_axis({'line': {'none': True}})
        graph.set_legend({'none': True})  # 设置图例
        worksheet.insert_chart(start_row + index * 7, end_column + 1, graph)

        # WoE
        column_graph = workbook.add_chart({'type': 'column'})
        range_column_index = df.columns.get_loc('range') + start_column
        badrate_column_index = df.columns.get_loc('woe') + start_column

        range_categories = [worksheet.get_name(), start_row, range_column_index, end_row - 1,
                            range_column_index]
        range_value = [worksheet.get_name(), start_row, badrate_column_index, end_row - 1,
                       badrate_column_index]
        column_graph.add_series({'name': 'woe', 'categories': range_categories, 'values': range_value,
                                 'marker': {'type': 'circle'}, 'data_labels': {'value': True}})
        column_graph.set_size({'width': 450, 'height': 200})  # TODO 大小改成和单元格的大小一致,以单元格的大小为单位
        column_graph.set_title(
            {'name': 'WoE - {}'.format(feature), 'overlay': False,
             'name_font': {'name': '微软雅黑', 'size': 9, 'bold': True}})
        column_graph.set_x_axis({'line': {'none': True}, 'label_position': 'low'})
        column_graph.set_y_axis({'line': {'none': True}})
        column_graph.set_legend({'none': True})  # 设置图例
        worksheet.insert_chart(start_row + index * 7, end_column + 9, column_graph)

        # 给变量名称插入跳转超链接
        for i in range(1, df.shape[0] + 1):
            worksheet.write_url('B{}'.format(start_row + i),
                                'internal:{}!M{}:AC{}'.format(worksheet.get_name(), start_row + index * 7 + 1,
                                                              start_row + index * 7 + 10), string=feature,
                                tip='Jump to charts')

    # 写入标题
    for col_num, value in enumerate(var_summary_df.columns.values):
        worksheet.write(0, col_num, value, header_format)

    # 写入数据和插入图像
    start_row, start_column = 1, 0
    index = 0
    for name, single_df in var_summary_df.groupby('var_name'):
        write_df_to_excel_by_row(index, name, single_df, workbook, worksheet, start_row, start_column)
        start_row = single_df.shape[0] + start_row
        index += 1

    # 冻结窗格
    worksheet.freeze_panes(1, 2)
    worksheet.set_column(0, 1, 18)

    # # 保存为excel
    # workbook.close()


if __name__ == "__main__":
    import pandas as pd
    import numpy as np

    # 创建一个数据
    df = pd.DataFrame(np.random.randn(182, 9), columns=list('ABCDEFGHI'))
    column_list = df.columns
    # 使用XlsxWriter引擎创建一个pandas Excel writer。
    writer = pd.ExcelWriter("../tests/test_report2excel.xlsx", engine='xlsxwriter')

    df.to_excel(writer, index=False)

    workbook = writer.book
    worksheet = writer.sheets['Sheet1']
    ###########################################
    # worksheet.set_landscape()
    # worksheet.set_paper(8)
    # worksheet.set_margins(0.787402, 0.787402, 0.5, 0.787402)

    apply_outer_border_to_range(
        workbook,
        worksheet,
        {
            "range_string": "C10:M20",
            "border_style": 5,
        },
    )

    # 关闭workbook
    workbook.close()



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: category_label_encoder.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-09-21
'''

# import pandas as pd
# from sklearn.preprocessing import LabelEncoder
#
#
# def category_to_labelencoder(data, labelencoder=[]):
#     label_encoder_dict = {}
#     le = LabelEncoder()
#     for col in labelencoder:
#         print('{} in process!!!'.format(col))
#         data[col] = le.fit_transform(data[col].values)
#         number = [i for i in range(0, len(le.classes_))]
#         key = list(le.inverse_transform(number))
#         label_encoder_dict[col] = dict(zip(key, number))
#     return label_encoder_dict
#
#
# def category_to_labelencoder_apply(data, labelencoder_dict={}):
#     for col, mapping in labelencoder_dict.items():
#         print('{} in process!!!'.format(col))
#         data[col] = data[col].map(mapping).fillna(-1)
#         data[col] = data[col].astype(int)
#
#
# #####训练
# fruit_data = pd.DataFrame({
#     'fruit': ['apple', 'orange', 'pear', 'orange', 'red'],
#     'color': ['red', 'orange', 'green', 'green', 'red'],
#     'weight': [5, 6, 3, 4, 2]
# })
# print(fruit_data)
#
# labelencoder_cols = ['fruit', 'color']
#
# label_encoder_dict = category_to_labelencoder(fruit_data, labelencoder_cols)
# print(fruit_data)
#
# #####应用
# test_data = pd.DataFrame({
#     'fruit': ['apple', 'orange', 'pear', 'orange', 'red'],
#     'color': ['aaa', 'orange', 'green', 'green', 'red'],
#     'weight': [5, 6, 3, 4, 2]
# })
# print(test_data)
#
# category_to_labelencoder_apply(test_data, label_encoder_dict)
# print(test_data)


print('########################'*5)

import pandas as pd

df = pd.DataFrame([
    ['green', 'Chevrolet', 2017],
    ['blue', 'BMW', 2015],
    ['yellow', 'Lexus', 2018],
])
df.columns = ['color', 'make', 'year']
df.to_csv('df_source.csv')

df_processed = pd.get_dummies(df, prefix_sep="_", columns=df.columns[:-1])
print(df_processed.head(10))
df_processed.to_csv('df_processed.csv')

aa = ['aa', 'bb', 'cc']
print(aa)
aa.remove('aa')
print(aa)

df = pd.DataFrame([
    ['green;yellow;aaa', 'Chevrolet', 2017],
    ['blue;green', 'BMW', 2015],
    ['yellow', 'Lexus', 2018],
])
df.columns = ['color', 'make', 'year']
print(df)

print(df.set_index(["make", "year"])["color"].str.split(";", expand=True))
df = df.set_index(["make", "year"])["color"].str.split(";", expand=True).reset_index()
print(df)
del df['make']
print(pd.get_dummies(df,prefix='', prefix_sep='').groupby(level=0, axis=1).max())
###https://www.cnblogs.com/traditional/p/11967360.html?ivk_sa=1024320u
###https://www.5axxw.com/questions/content/s2rhn2


df = pd.DataFrame([
    ['green;yellow;aaa', 'Chevrolet;Lexus', 2017],
    ['blue;green', 'BMW', 2015],
    ['yellow', 'Lexus;BMW', 2018],
])
df.columns = ['color', 'make', 'year']
print(df)

print(df.set_index(["year"])["color"].str.split(";", expand=True))
df = df.set_index(["year"])["color"].str.split(";", expand=True).reset_index()
print(df)
print(pd.get_dummies(df,prefix='', prefix_sep='').groupby(level=0, axis=1).max())
exit(0)
# df = df.set_index(["make", "year"])["color"].str.split(";", expand=True).stack().reset_index(drop=True, level=-1).reset_index().rename(columns={0: "color"})
# print(df)
# del df['make']
# print(df)
# print(pd.get_dummies(df,prefix='', prefix_sep='').groupby(level=0, axis=1).max().set_index("year"))

# df_processed = pd.get_dummies(df, prefix_sep="_", columns=['color'])
# print(df_processed.head(10))
# df_processed.to_csv('df_processed_get_dummies.csv')


df = pd.DataFrame([
    [101, 'roof', 'garage', 'basement'],
    [102, 'basement', 'garage', 'painting'],
])
df.columns = ['no', 'point1', 'point2', 'point3']
print(df)

#print(pd.get_dummies(df))
print(pd.get_dummies(df,prefix='', prefix_sep='').groupby(level=0, axis=1).max().set_index("no"))


#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: logger_utils.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-06-06
'''

import logging


# 日志输出
class Logger():
    # 日志级别关系映射
    level_relations = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL
    }

    def __init__(self, level="info", name=None,
                 fmt="%(asctime)s - %(name)s[line:%(lineno)d] - %"
                     "(levelname)s: %(message)s"):
        logging.basicConfig(level=self.level_relations.get(level), format=fmt)
        self.logger = logging.getLogger(name)



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-


import os
import sys
module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path)

import numpy as np
import pandas as pd
import pytest

from autobmt.metrics import get_auc, get_ks, psi

np.random.seed(1)

feature = np.random.rand(500)
target = np.random.randint(2, size=500)
base_feature = np.random.rand(500)

test_df = pd.DataFrame({
    'A': np.random.rand(500),
    'B': np.random.rand(500),
})
base_df = pd.DataFrame({
    'A': np.random.rand(500),
    'B': np.random.rand(500),
})

FUZZ_THRESHOLD = 1e-10


def test_get_ks():
    result = get_ks(target, feature)
    assert result == 0.055367756612569874


def test_get_auc():
    result = get_auc(target, feature)
    assert result == 0.5038690142424582


def test_PSI():
    result = psi(feature, base_feature, featurebin=[0.3, 0.5, 0.7])
    assert result == 0.018630024627491467


def test_PSI_frame():
    result = psi(
        test_df,
        base_df,
        featurebin={
            'A': [0.3, 0.5, 0.7],
            'B': [0.4, 0.8],
        },
    )

    assert result['B'] == pytest.approx(0.014528279995858708, FUZZ_THRESHOLD)


def test_PSI_return_frame():
    result, frame = psi(
        test_df,
        base_df,
        featurebin={
            'A': [0.3, 0.5, 0.7],
            'B': [0.4, 0.8],
        },
        return_frame=True,
    )

    assert frame.loc[4, 'no_base'] == 0.38



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-


import os
import sys
module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path)

import numpy as np
import pandas as pd

from autobmt.feature_selection import select_features_by_miss, select_features_by_concentrate, \
    select_features_by_iv_diff, select_features_by_iv, select_features_by_corr, select_features_by_psi

np.random.seed(1)

LENGTH = 500

A = np.random.rand(LENGTH)
A[np.random.choice(LENGTH, 20, replace=False)] = np.nan

B = np.random.randint(100, size=LENGTH)
C = A + np.random.normal(0, 0.2, LENGTH)
D = A + np.random.normal(0, 0.1, LENGTH)

E = np.random.rand(LENGTH)
E[np.random.choice(LENGTH, 480, replace=False)] = np.nan

F = B + np.random.normal(0, 10, LENGTH)

target = np.random.randint(2, size=LENGTH)

frame = pd.DataFrame({
    'A': A,
    'B': B,
    'C': C,
    'D': D,
    'E': E,
    'F': F,
})

noframe = pd.DataFrame({
    'A': np.random.shuffle(A),
    'B': B,
    'C': C,
    'D': D,
    'E': E,
    'F': F,
})

frame['target'] = target
noframe['target'] = target


def test_select_features_by_miss():
    df = select_features_by_miss(frame, threshold=0.8)
    assert 'E' not in df


def test_select_features_by_concentrate():
    df = select_features_by_concentrate(frame, threshold=0.2)
    # assert 'target' not in df


def test_select_features_by_psi():
    df = select_features_by_psi(frame, noframe, threshold=0.05)
    print("select_features_by_psi", df.columns)
    assert 'A' not in df


def test_select_features_by_iv():
    df = select_features_by_iv(frame, target='target', threshold=0.25)
    assert 'B' not in df


def test_select_features_by_iv_diff():
    df = select_features_by_iv_diff(frame, noframe, target='target', threshold=2)
    assert 'A' not in df


def test_select_features_by_corr():
    df = select_features_by_corr(frame, target='target')
    assert ['D', 'E', 'F', 'target'] == df.columns.tolist()

# import time
#
# import numpy as np
# import pandas as pd
#
# cust_id = 'gid'
# target = 'y_label'
# apply_time = 'apply_time'
# data_type = 'type'
#
# time_start = time.time()
#
# exclude_cols = [cust_id, target, data_type]
#
# train_data = pd.read_csv('./ym_train_data.csv')
# oot_data = pd.read_csv('./ym_test_data.csv')
#
# train_data.replace({-999999: np.nan}, inplace=True)
# oot_data.replace({-999999: np.nan}, inplace=True)
#
# train_data['type'] = 'train'
# oot_data['type'] = 'test'
#
# all_data = train_data.append(oot_data)
#
# merge_col = list(train_data.columns)
# merge_col.remove(target)
# merge_col.remove(cust_id)
# merge_col.remove(data_type)
#
# from .transformer import FeatureBin
# from .feature_selection import select_features_by_miss, select_features_by_concentrate, select_features_by_iv, \
#     select_features_by_iv_diff, select_features_by_corr, select_features_by_psi

# ==============================单个单个测试

# # select_features_by_miss
# drop_df, drop_cols = select_features_by_miss(train_data, threshold=0.9, include_cols=merge_col, return_drop=True)
# print(len(drop_cols))
# print("剔除的变量")
# print(drop_cols)
# print(drop_df.shape)
#
# # select_features_by_concentrate
# drop_df, drop_cols = select_features_by_concentrate(train_data, threshold=0.5, include_cols=merge_col, return_drop=True)
# print(len(drop_cols))
# print("剔除的变量")
# print(drop_cols)
# print(drop_df.shape)

# # feature_bin not None
# fb = FeatureBin()
# # select_features_by_iv_diff
# drop_df, drop_cols, iv = select_features_by_iv_diff(train_data, oot_data, target=target, threshold=2, include_cols=merge_col, return_drop=True, return_iv=True, feature_bin=fb)
# print(len(drop_cols))
# print("剔除的变量")
# print(drop_cols)
# iv.to_csv('../tests/feature_selection_test_py_select_features_by_iv_diff____iv_calciv.csv')
# print(drop_df.shape)

# select_features_by_iv
# drop_df, drop_cols, iv = select_features_by_iv(train_data, target=target, threshold=0.02, include_cols=merge_col, return_drop=True, return_iv=True, feature_bin=fb)
# drop_df, drop_cols, iv = select_features_by_iv(train_data, target=target, threshold=0.02, include_cols=merge_col, return_drop=True, return_iv=True)
# print(len(drop_cols))
# print("剔除的变量")
# print(drop_cols)
# #iv.to_csv('../tests/feature_selection_test_py_select_features_by_iv____iv_calciv_check.csv')
# iv.to_csv('../tests/feature_selection_test_py_select_features_by_iv____iv_calciv_check_update.csv')
# print(drop_df.shape)

# # select_features_by_iv_diff
# drop_df, drop_cols, iv = select_features_by_iv_diff(train_data, oot_data, target=target, threshold=2,include_cols=merge_col, return_drop=True, return_iv=True)
# print(len(drop_cols))
# print("剔除的变量")
# print(drop_cols)
# iv.to_csv('../tests/feature_selection_test_py_select_features_by_iv_diff____iv.csv')
# print(drop_df.shape)
#
# # select_features_by_corr
# drop_df, drop_cols, iv = select_features_by_corr(train_data, target=target, threshold=0.7, include_cols=merge_col,return_drop=True)
# print(len(drop_cols))
# print("剔除的变量")
# print(drop_cols)
# iv.to_csv('../tests/feature_selection_test_py_select_features_by_corr____iv.csv')
# print(drop_df.shape)
#
# #select_features_by_psi
# drop_df, drop_cols, psi = select_features_by_psi(train_data, oot_data, target=target, threshold=0.05,include_cols=merge_col, return_drop=True, feature_bin=True)
# print(len(drop_cols))
# print("剔除的变量")
# print(drop_cols)
# psi.to_csv('../tests/feature_select_features_by_psi____psi.csv')
# print(drop_df.shape)

# ==============================单个单个测试

# #=================删除指定变量，测试相关性
# del_c = ['E138', 'D033', 'E092', 'D159', 'E086', 'E137', 'D064', 'D083', 'D058', 'E128', 'D035', 'D165', 'E084', 'D152', 'D008', 'D012', 'D170', 'D049', 'D024', 'D009', 'D084', 'D132', 'D110', 'D025', 'E125', 'E136', 'D034', 'E150', 'D088', 'D108', 'D129', 'D037', 'D102', 'D169', 'D109', 'D061', 'D161', 'E116', 'D174', 'D013', 'E104', 'D010', 'D133', 'D082', 'E101', 'D153', 'D079', 'E122', 'D085', 'E087', 'E102', 'E093', 'D011', 'D078', 'D036', 'E131', 'D059', 'D032', 'D128', 'E106', 'D060', 'D173', 'D103', 'E085', 'D134', 'D130', 'E132']
# # select_features_by_corr
# drop_df, drop_cols, iv = select_features_by_corr(train_data.drop(columns=del_c), target=target, threshold=0.7, include_cols=list(set(merge_col)-set(del_c)),return_drop=True)
# print(len(drop_cols))
# print("剔除的变量")
# print(drop_cols)
# iv.to_csv('../tests/feature_selection_test_py_select_features_by_corr____iv删除特定变量后的.csv')
# print(drop_df.shape)


# #==============================流程测试
# from ryan.feature_selection import FeatureSelection
# #from ryan.feature_selection_v1 import FeatureSelection
# #===psi优先
# # fs_dic = {
# #         "psi": {'threshold': 0.05},
# #         "empty": {'threshold': 0.9},
# #         "const": {'threshold': 0.5},
# #         "iv": {'threshold': 0.02},
# #         "iv_diff": {'threshold': 2, 'target': target},
# #         "corr": {'threshold': 0.7},
# #
# #     }
# #
# # fs_dic = {
# #         "iv": {'threshold': 0.02},
# #         "psi": {'threshold': 0.05},
# #         "empty": {'threshold': 0.9},
# #         "const": {'threshold': 0.5},
# #         "iv_diff": {'threshold': 2, 'target': target},
# #         "corr": {'threshold': 0.7},
# #
# #     }
# #
# # fs_dic = {
# #         "iv_diff": {'threshold': 2, 'target': target},
# #         "empty": {'threshold': 0.9},
# #         "const": {'threshold': 0.5},
# #         "iv": {'threshold': 0.02},
# #         "psi": {'threshold': 0.05},
# #         "corr": {'threshold': 0.7},
# #
# #     }
# #
# # fs_dic = {
# #         "corr": {'threshold': 0.7},
# #         "empty": {'threshold': 0.9},
# #         "const": {'threshold': 0.5},
# #         "iv": {'threshold': 0.02},
# #         "psi": {'threshold': 0.05},
# #         "iv_diff": {'threshold': 2, 'target': target},
# #
# #     }
# #
# # fs_dic = {
# #         "empty": {'threshold': 0.9},
# #         "const": {'threshold': 0.5},
# #         "iv": {'threshold': 0.02},
# #         "psi": {'threshold': 0.05},
# #         "iv_diff": {'threshold': 2, 'target': target},
# #         "corr": {'threshold': 0.7},
# #
# #     }
# #
# # fs_dic = {
# #         "const": {'threshold': 0.5},
# #         "empty": {'threshold': 0.9},
# #         "iv": {'threshold': 0.02},
# #         "psi": {'threshold': 0.05},
# #         "iv_diff": {'threshold': 2, 'target': target},
# #         "corr": {'threshold': 0.7},
# #
# #     }
#
# fs_dic = {
#         "psi": {'threshold': 0.05},
#         "empty": {'threshold': 0.9},
#         "const": {'threshold': 0.5},
#         "corr": {'threshold': 0.7},
#         "iv_diff": {'threshold': 2},
#         "iv": {'threshold': 0.02},
#
#     }
#
#
# print("===============fs_dicfs_dic===============")
#
# fs = FeatureSelection(df=all_data, target=target, exclude_columns=exclude_cols, match_dict=None,
#                       params=fs_dic)
# single_df, selected_features, select_log_df, selection_evaluate_log_df = fs.select()
# print(selected_features)
# print(len(selected_features)) #82
# print(select_log_df.head(3))
#
# #select_log_df.to_csv('../tests/feature_selection_test_psinot2bin_psi优先_selection.csv', index=False)
# #select_log_df.to_csv('../tests/feature_selection_test_psinot2bin_iv优先_selection.csv', index=False)
# #select_log_df.to_csv('../tests/feature_selection_test_psinot2bin_ivdiff优先_selection.csv', index=False)
# #select_log_df.to_csv('../tests/feature_selection_test_psinot2bin_corr优先_selection.csv', index=False)
# #select_log_df.to_csv('../tests/feature_selection_test_psinot2bin_empty优先_selection.csv', index=False)
# #select_log_df.to_csv('../tests/feature_selection_test_psinot2bin_const优先_selection.csv', index=False)
# select_log_df.to_csv('../tests/feature_selection_test_psinot2bin_psi优先ivivdiff_selection.csv', index=False)
#
#
# # #==============================流程测试无指定筛选方式和阀值筛选
# # fs = FeatureSelection(df=all_data, target=target, exclude_columns=exclude_cols, match_dict=None,)
# # single_df, selected_features, select_log_df, selection_evaluate_log_df = fs.select()
# # print(selected_features)
# # print(len(selected_features)) #82
# # print(select_log_df.head(3))
# # select_log_df.to_csv('../tests/feature_selection_test_psinot2bin_无params.csv', index=False)
#
#
#
# # selection_evaluate_log_df.to_csv('selection_evaluate_log_df_new.csv', index=False)
# time_end = time.time()
# time_c = time_end - time_start
# print('time cost {} s'.format(time_c))
# # time cost 6.1279308795928955 s
# ####################################new
# import sys
#
# sys.exit(0)



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: metrics.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-11-05
'''

import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score

from .utils import del_df, unpack_tuple


# auc
def get_auc(target, y_pred):
    """
    计算auc值
    Args:
        target (array-like): 目标变量列表
        y_pred (array-like): 模型预测的分数或概率列表

    Returns:
        float: auc值
    """
    if len(np.unique(target)) != 2:
        raise ValueError('the target is not 2 classier target')
    else:
        return roc_auc_score(target, y_pred)


# ks
def get_ks(target, y_pred):
    """
    计算ks值
    Args:
        target (array-like): 目标变量列表
        y_pred (array-like): 模型预测的分数或概率列表

    Returns:
        float: ks值
    """
    df = pd.DataFrame({
        'y_pred': y_pred,
        'target': target,
    })
    crossfreq = pd.crosstab(df['y_pred'], df['target'])
    crossdens = crossfreq.cumsum(axis=0) / crossfreq.sum()
    crossdens['ks'] = abs(crossdens[0] - crossdens[1])
    ks = max(crossdens['ks'])
    return ks


def get_auc_ks_psi(df, target='target', pred='p'):
    """
    计算auc、ks、psi
    Args:
        df:
        target:
        pred:

    Returns:

    """
    df_pred_dev = df[df['type'] == 'train']
    df_pred_nodev = df[df['type'] == 'test']

    # 计算auc、ks、psi
    test_ks = get_ks(df_pred_nodev[target], df_pred_nodev[pred])
    train_ks = get_ks(df_pred_dev[target], df_pred_dev[pred])
    test_auc = get_auc(df_pred_nodev[target], df_pred_nodev[pred])
    train_auc = get_auc(df_pred_dev[target], df_pred_dev[pred])

    q_cut_list = np.arange(0, 1, 1 / 20)
    bins = np.append(np.unique(np.quantile(df_pred_nodev[pred], q_cut_list)), df_pred_nodev[pred].max() + 0.1)
    df_pred_nodev['range'] = pd.cut(df_pred_nodev[pred], bins=bins, precision=0, right=False).astype(str)
    df_pred_dev['range'] = pd.cut(df_pred_dev[pred], bins=bins, precision=0, right=False).astype(str)
    nodev_psi = psi(df_pred_nodev['range'], df_pred_dev['range'])
    res_dict = {'dev_auc': train_auc, 'nodev_auc': test_auc, 'dev_ks': train_ks, 'nodev_ks': test_ks,
                'nodev_dev_psi': nodev_psi}
    del_df(df_pred_dev)
    del_df(df_pred_nodev)
    return pd.DataFrame([res_dict])


def psi(no_base, base, return_frame=False, featurebin = None):
    """
    计算psi值
    Args:
        no_base (DataFrame|array-like): 非基准数据集
        base (DataFrame|array-like):基准数据集
        return_frame (bool): 是否需要返回占比

    Returns:
        float|Series:psi值
    """

    if featurebin is not None:
        if isinstance(featurebin, (dict, list)):
            from .transformer import FeatureBin
            featurebin = FeatureBin().manual_bin(featurebin)

            no_base = featurebin.transform(no_base, labels = True)
        base = featurebin.transform(base, labels = True)

    psi = list()
    frame = list()

    if isinstance(no_base, pd.DataFrame):
        for col in no_base:
            p, f = calc_psi(no_base[col], base[col])
            psi.append(p)
            frame.append(f)

        psi = pd.Series(psi, index=no_base.columns, name='psi')

        frame = pd.concat(
            frame,
            keys=no_base.columns,
            names=['columns', 'id'],
        ).reset_index()
        frame = frame.drop(columns='id')
    else:
        psi, frame = calc_psi(no_base, base)

    res = (psi,)

    if return_frame:
        res += (frame,)

    return unpack_tuple(res)


def calc_psi(no_base, base):
    """
    psi计算的具体逻辑
    Args:
        no_base (array-like): 非基准数据集
        base (array-like): 基准数据集

    Returns:
        float,DataFrame : psi值，占比
    """
    no_base_prop = pd.Series(no_base).value_counts(normalize=True, dropna=False)
    base_prop = pd.Series(base).value_counts(normalize=True, dropna=False)

    psi = np.sum((no_base_prop - base_prop) * np.log(no_base_prop / base_prop))

    frame = pd.DataFrame({
        'no_base': no_base_prop,
        'base': base_prop,
    })
    frame.index.name = 'value'

    return psi, frame.reset_index()


def SSE(y_pred, y):
    """sse
    """
    return np.sum((y_pred - y) ** 2)


def MSE(y_pred, y):
    """mse
    """
    return np.mean((y_pred - y) ** 2)


def AIC(y_pred, y, k, llf=None):
    """AIC信息准则

    Args:
        y_pred (array-like)
        y (array-like)
        k (int): x的数量
        llf (float): 对数似然函数的值
    """
    if llf is None:
        llf = np.log(SSE(y_pred, y))

    return 2 * k - 2 * llf


def BIC(y_pred, y, k, llf=None):
    """贝叶斯信息准则

    Args:
        y_pred (array-like)
        y (array-like)
        k (int): x的数量
        llf (float): 对数似然函数的值
    """
    n = len(y)
    if llf is None:
        llf = np.log(SSE(y_pred, y))

    return np.log(n) * k - 2 * llf



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: __init__.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2022-07-09
'''

from .feature_binning import bin_method_run, chi_bin, dt_bin, equal_freq_bin, kmeans_bin, best_binning
from .detector import detect
from .feature_selection import FeatureSelection
from .report2excel import Report2Excel, var_summary_to_excel
from .statistics import calc_bin_summary, calc_var_summary, compare_inflection_point, get_vif
from .stepwise import stepwise, StatsModel
from .transformer import WoeTransformer, FeatureBin
from .metrics import psi, get_auc_ks_psi, get_auc, get_ks
from .utils import del_df, dump_to_pkl, load_from_pkl, fea_woe_dict_format, to_score
from .logger_utils import Logger
from .scorecard import ScoreCard
from .plot import plot_var_bin_summary


__version__ = "0.1.1"
VERSION = __version__




#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: stepwise.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-03-06
'''
import warnings

import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy import stats
from sklearn.linear_model import LogisticRegression

from .logger_utils import Logger
from .metrics import get_auc, get_ks, AIC, BIC, MSE
from .utils import split_target, unpack_tuple, step_evaluate_models, \
    get_max_corr_feature, model_predict_evaluate

INTERCEPT_COLS = 'intercept'

warnings.filterwarnings(action='ignore')
pd.set_option('display.max_columns', None)
log = Logger(level="info", name=__name__).logger


class StatsModel:
    def __init__(self, estimator='ols', criterion='aic', intercept=False):
        if isinstance(estimator, str):
            Est = self.get_estimator(estimator)
            estimator = Est(fit_intercept=intercept, )

        self.estimator = estimator
        self.intercept = intercept
        self.criterion = criterion

    def get_estimator(self, name):
        from sklearn.linear_model import (
            LinearRegression,
            LogisticRegression,
            Lasso,
            Ridge,
        )

        ests = {
            'ols': LinearRegression,
            'lr': LogisticRegression,
            'lasso': Lasso,
            'ridge': Ridge,
        }

        if name in ests:
            return ests[name]

        raise Exception('estimator {name} is not supported'.format(name=name))

    def stats(self, X, y):
        """
        """
        X = X.copy()

        if isinstance(X, pd.Series):
            X = X.to_frame()

        self.estimator.fit(X, y)

        if hasattr(self.estimator, 'predict_proba'):
            pre = self.estimator.predict_proba(X)[:, 1]
        else:
            pre = self.estimator.predict(X)

        coef = self.estimator.coef_.reshape(-1)

        if self.intercept:
            coef = np.append(coef, self.estimator.intercept_)
            X[INTERCEPT_COLS] = np.ones(X.shape[0])

        n, k = X.shape

        t_value = self.t_value(pre, y, X, coef)
        p_value = self.p_value(t_value, n)
        c = self.get_criterion(pre, y, k)

        return {
            't_value': pd.Series(t_value, index=X.columns),
            'p_value': pd.Series(p_value, index=X.columns),
            'criterion': c
        }

    def get_criterion(self, pre, y, k):
        if self.criterion == 'aic':
            llf = self.loglikelihood(pre, y, k)
            return AIC(pre, y, k, llf=llf)

        if self.criterion == 'bic':
            llf = self.loglikelihood(pre, y, k)
            return BIC(pre, y, k, llf=llf)

        if self.criterion == 'ks':
            return get_ks(y, pre)

        if self.criterion == 'auc':
            return get_auc(y, pre)

    def t_value(self, pre, y, X, coef):
        n, k = X.shape
        mse = sum((y - pre) ** 2) / float(n - k)
        nx = np.dot(X.T, X)

        if np.linalg.det(nx) == 0:
            return np.nan

        std_e = np.sqrt(mse * (np.linalg.inv(nx).diagonal()))
        return coef / std_e

    def p_value(self, t, n):
        return stats.t.sf(np.abs(t), n - 1) * 2

    def loglikelihood(self, pre, y, k):
        n = len(y)
        mse = MSE(pre, y)
        return (-n / 2) * np.log(2 * np.pi * mse * np.e)


def stepwise(frame, target='target', estimator='ols', direction='both', criterion='aic',
             p_enter=0.01, p_remove=0.01, p_value_enter=0.2, intercept=False,
             max_iter=None, return_drop=False, exclude=None):
    """
    逐步回归选择特征
    Args:
        frame (DataFrame): 用于训练模型的数据集
        target (str): 目标变量名称
        estimator (str): 用于统计的模型
        direction (str): 前向逐步还是后向逐步, 支持“forward”、“backward”和“both”，建议“both”
        criterion (str): 统计模型的信息准则, 支持“aic”、“bic”
        p_enter (float): 阈值将在“forward”和“both”中使用，用于保留特征
        p_remove (float): 阈值将在“backward”中使用，用于剔除特征
        intercept (bool): 是否需要截距项
        p_value_enter (float): 阈值将在“both”中使用，用于剔除特征
        max_iter (int): 最大迭代次数
        return_drop (bool): 是否需要返回删除的特征
        exclude (array-like): 不参与特征筛选的特征列表

    Returns:
        DataFrame: 筛选后的数据集
        array: 删除的特征列表
    """
    df, y = split_target(frame, target)

    if exclude is not None:
        df = df.drop(columns=list(set(exclude) & set(df.columns)))

    drop_list = []
    remaining = df.columns.tolist()

    selected = []

    sm = StatsModel(estimator=estimator, criterion=criterion, intercept=intercept)

    order = -1 if criterion in ['aic', 'bic'] else 1

    best_score = -np.inf * order

    iter = -1
    while remaining:
        iter += 1
        if max_iter and iter > max_iter:
            break

        l = len(remaining)
        test_score = np.zeros(l)
        test_res = np.empty(l, dtype=object)

        if direction == 'backward':
            for i in range(l):
                test_res[i] = sm.stats(
                    df[remaining[:i] + remaining[i + 1:]],
                    y,
                )
                test_score[i] = test_res[i]['criterion']

            curr_ix = np.argmax(test_score * order)
            curr_score = test_score[curr_ix]

            if (curr_score - best_score) * order < p_remove:
                break

            name = remaining.pop(curr_ix)
            drop_list.append(name)

            best_score = curr_score

        # forward and both
        else:
            for i in range(l):
                test_res[i] = sm.stats(
                    df[selected + [remaining[i]]],
                    y,
                )
                test_score[i] = test_res[i]['criterion']

            curr_ix = np.argmax(test_score * order)
            curr_score = test_score[curr_ix]

            name = remaining.pop(curr_ix)
            if (curr_score - best_score) * order < p_enter:
                drop_list.append(name)

                # early stop
                if selected:
                    drop_list += remaining
                    break

                continue

            selected.append(name)
            best_score = curr_score

            if direction == 'both':
                p_values = test_res[curr_ix]['p_value']
                drop_names = p_values[p_values > p_value_enter].index

                for name in drop_names:
                    selected.remove(name)
                    drop_list.append(name)

    r = frame.drop(columns=drop_list)

    res = (r,)
    if return_drop:
        res += (drop_list,)

    return unpack_tuple(res)


class StepWise:
    def __init__(self, df, target='target', features=[], exclude_columns=None, iv_psi_df=None, var_bin_woe={},
                 match_dict={},
                 object='test_ks', features_corr_df=None,
                 include_columns=[], p_threshold=0.05, vif_threshold=5, max_varcnt=10, A=404.65547022, B=72.1347520444,
                 is_return_var=False):
        self.df = df
        self.target = target
        self.features = features
        self.exclude_columns = exclude_columns
        self.iv_psi_df = iv_psi_df
        self.var_bin_woe = var_bin_woe
        self.match_dict = match_dict  # 特征的中文字典名称
        self.object = object
        self.features_corr_df = features_corr_df  # 变量相关性替换列表
        self.include_columns = include_columns
        self.p_threshold = p_threshold
        self.vif_threshold = vif_threshold
        self.A = A
        self.B = B
        self.max_varcnt = len(self.features) if max_varcnt > len(self.features) else max_varcnt
        self.step_evaluate_log_df = []
        self.predict_data = None
        self.is_return_var = is_return_var

    @property
    def get_evaluate_df_log(self):
        """合并每一步的评估结果"""
        if len(self.step_evaluate_log_df) == 0:
            log.info("并未进行评估过!!!")
            return None
        else:
            evaluate_log_df = pd.concat(self.step_evaluate_log_df, axis=0).reset_index(drop=True)
            return evaluate_log_df

    @property
    def get_output_data(self):
        """获取预测的结果"""
        if self.predict_data is not None and isinstance(self.predict_data, pd.DataFrame):
            if self.is_return_var:
                for k, v in self.model.params.to_dict().items():
                    if k == 'const':
                        self.predict_data['const'] = round(self.A - self.B * v)
                    else:
                        self.predict_data["{}_fscore".format(k)] = self.predict_data[k].map(
                            lambda x: round(-(self.B * v * x)))

                # dump_to_pkl(score_card_structure,'./score_card_structure.pkl')
                in_model_features = [i for i in self.predict_data.columns.tolist() if '_fscore' in i] + ['const']
                self.predict_data['woe2score'] = self.predict_data[in_model_features].sum(axis=1).map(
                    lambda x: round(x))
                return self.predict_data
            else:
                return self.predict_data
        else:
            ValueError("并没有进行预测，请先执行stepwise_apply()!")

    def run(self):
        """
        新版双向stepwise方法
        Returns:

        """
        # 执行了最优分箱后auc,ks
        step_name = "best_binning"
        step_evaluate_df, evaluate_log = step_evaluate_models(self.df, self.features, self.target, stepname=step_name)
        self.step_evaluate_log_df.append(step_evaluate_df)
        log.info("{},evaluate --- {}".format(step_name, evaluate_log))

        log.info("开始进行Stepwise,变量个数为:{}".format(len(self.features)))
        # 前置检查
        self.check()
        log.info("Stepwise前置检查完成")
        record_features_log = []  # 记录整个stepwise的变量

        initial_list = self.include_columns.copy()  # 保留的特征
        train_data = self.df[self.df['type'] == 'train']
        test_data = self.df[self.df['type'] == 'test']
        if self.iv_psi_df is not None and isinstance(self.iv_psi_df, pd.DataFrame):
            self.features = self.iv_psi_df.sort_values(by='IV', ascending=False).index.to_list()
        high_iv_name = self.features[0]
        initial_list.append(high_iv_name)
        self.features.remove(high_iv_name)
        flag_features = initial_list.copy()  # 记录特征
        # 最大iv的目标值，作为基准
        base_object_value, _ = get_object_value_by_train_lr(train_data[initial_list], train_data[self.target],
                                                            test_data[initial_list], test_data[self.target])
        record_features_log.append(("round-0", ",".join(initial_list), base_object_value, 0))
        log.info("最高iv变量{}的基础目标值为{}".format(high_iv_name, base_object_value))
        step = 1
        while True:
            result_dict = {}
            # 前向选择过程
            # 遍历判断添加一个变量后，目标值是否有提升
            for name in self.features:
                if name not in flag_features:
                    object_value, _ = get_object_value_by_train_lr(train_data[initial_list + [name]],
                                                                   train_data[self.target],
                                                                   test_data[initial_list + [name]],
                                                                   test_data[self.target])
                    result_dict[name] = object_value
            max_key, max_value = get_max_value_in_dic(result_dict)  # 返回最优的特征和object值
            # 若有提升，则加入到候选变量中
            round_history_arr = [("rount-{}".format(step), ",".join(initial_list + [k]), v, v - base_object_value) for
                                 k, v in
                                 result_dict.items()]
            record_features_log.extend(round_history_arr)

            if max_value > base_object_value:
                initial_list.append(max_key)  # 加入到候选变量
                flag_features.append(max_key)  # 记录该特征已经选择过
                base_object_value = max_value  # 更新当前最优object值
                # 后向判断过程
                df_pvalue_vif, initial_list = calculate_features_p_value_vif(train_data, initial_list, self.target)
                filter_df = df_pvalue_vif[(df_pvalue_vif['pvalue'] > self.p_threshold) | (
                    df_pvalue_vif['vif'] > self.vif_threshold)]
                if filter_df.shape[0] > 0:
                    drop_col = filter_df.index.tolist()
                    # 删除后向剔除的特征
                    # TODO
                    initial_list = [i for i in initial_list if i not in drop_col]
                    drop_log = ("rount-{}-drop".format(step), ",".join(drop_col), 0, 0)  # 被剔除变量
                    record_features_log.append(drop_log)
                tmp = ("rount-{}-best".format(step), ",".join(initial_list), base_object_value, 0)  # 这一轮的最优变量
                record_features_log.append(tmp)

            step += 1
            # 若满足下列条件，则跳出循环
            if len(initial_list) >= self.max_varcnt or max_value - base_object_value < 0:
                break
        log.info("Stepwise目标选择完成,剩余变量个数为:{}".format(len(initial_list)))
        # 构造stepwise的过程输出和模型结果
        result_df, evaluate_df = self.output(train_data, initial_list, self.target)
        log_df = pd.DataFrame(record_features_log, columns=['round-n', 'features', 'object_value', 'diff'])
        step_name = "toad_stepwise"
        step_evaluate_df, evaluate_log = step_evaluate_models(self.df, initial_list, self.target, stepname=step_name)
        self.step_evaluate_log_df.append(step_evaluate_df)
        log.info("{},evaluate --- {}".format(step_name, evaluate_log))
        return initial_list, result_df, log_df, evaluate_df, self.get_evaluate_df_log

    def stepwise_apply(self):
        """
        基于pvalue、aic、bid的Stepwise方法
        Returns:

        """
        # 执行了最优分箱后auc,ks
        step_name = "best_binning"
        step_evaluate_df, evaluate_log = step_evaluate_models(self.df, self.features, self.target, stepname=step_name)
        self.step_evaluate_log_df.append(step_evaluate_df)
        log.info("{},evaluate --- {}".format(step_name, evaluate_log))

        train_data = self.df[self.df['type'] == 'train']
        if self.features_corr_df is not None and isinstance(self.features_corr_df, pd.DataFrame):
            max_corr_df = self.features_corr_df  # 如果相关性df传入了，则用传入的
        else:  # 否则用woe转换后的
            max_corr_df = get_max_corr_feature(train_data, self.features)  # 计算每个变量相关性最高的变量
        tmp_df = train_data[self.features + [self.target]]
        final_data = stepwise(tmp_df, target=self.target, estimator='ols', direction='both', criterion='aic',
                              p_enter=0.05, p_remove=0.05, )

        selected_features = [name for name in final_data.columns.to_list() if name not in self.exclude_columns]
        log.info("Stepwise目标选择完成,剩余变量个数为:{}".format(len(selected_features)))
        log.info("Stepwise保留变量为:{}".format(",".join(selected_features)))
        result_df, evaluate_df = self.output(train_data, selected_features, self.target)  # 训练模型
        step_name = "toad_stepwise"
        step_evaluate_df, evaluate_log = step_evaluate_models(self.df, selected_features, self.target,
                                                              stepname=step_name)
        self.step_evaluate_log_df.append(step_evaluate_df)
        log.info("{},evaluate --- {}".format(step_name, evaluate_log))

        # 关联上相关性最强的特征,方便替换
        result_df = result_df.merge(max_corr_df, how='left', left_index=True, right_index=True)
        if len(self.match_dict) > 0:
            result_df.insert(8, 'corr_name_cn', result_df['corr_name'].map(self.match_dict))
        result_df = result_df.reset_index()

        ##TODO 考虑在result_df进行psi、vif、相关性的过滤。result_df是10.model_stepwise

        return selected_features, result_df, evaluate_df, self.get_evaluate_df_log

    def output(self, train_data, features, target):
        '''

        Args:
            train_data:
            features:
            target:

        Returns:

        '''
        if features is None:
            raise ValueError("没有入模变量")

        # 训练模型,计算pvalue,vif
        df_pvalue_vif_coef, model, features = calculate_features_p_value_vif(train_data, features, target,
                                                                             need_coef=True,
                                                                             need_model=True)
        self.model = model
        if len(self.match_dict) > 0:
            df_pvalue_vif_coef.insert(0, 'cn', df_pvalue_vif_coef.index.map(self.match_dict))
        # 评估模型
        evaluate_df, self.predict_data = model_predict_evaluate(model, self.df, features, self.target, self.A, self.B,
                                                                self.exclude_columns, self.is_return_var)
        if self.iv_psi_df is not None:
            df_iv_psi = self.iv_psi_df[self.iv_psi_df.index.isin(features)]
            result_df = df_pvalue_vif_coef.merge(df_iv_psi, how='left', left_index=True, right_index=True)
            return result_df, evaluate_df
        else:
            return df_pvalue_vif_coef, evaluate_df

    def get_scorecard_structure(self):
        score_card_structure = {}
        for k, v in self.model.params.to_dict().items():
            # scorecard
            if k == 'const':
                score_card_structure['const'] = ['-', round(self.A - self.B * v)]
            else:
                one_feature_binning = self.var_bin_woe[k]
                one_feature_dict = {}
                for k1, v1 in one_feature_binning.items():
                    one_feature_dict[k1] = [float(v1), round(-(self.B * v * float(v1)))]  # v1是woe值
                score_card_structure[k] = one_feature_dict
        return score_card_structure

    def check(self):
        """
        特征选择模块，前置检查,符合要求，则往下运行
        Returns:

        """
        if len(self.features) <= 0 or self.features is None:
            raise ValueError("入模型变量不能为空，请检查!!!")
        if self.df is None or not isinstance(self.df, pd.DataFrame):
            raise ValueError("数据集不能为空并且数据集必须是dataframe!!!")
        if self.target is None:
            raise ValueError("数据集的目标变量名称不能为空!!!")
        if self.exclude_columns is None or "type" not in self.exclude_columns or self.target not in self.exclude_columns:
            raise ValueError("exclude 不能为空，必须包含target,type字段!!!")

    @staticmethod
    def scorecard_to_excel(scorecard, workbook, sheet_name):
        # excel相关格式
        # workbook = xlsxwriter.Workbook(output, {'nan_inf_to_errors': True})
        # worksheet = workbook.add_worksheet('分箱详情')
        worksheet = workbook.add_worksheet(sheet_name)
        # Add a header format.
        header_format = workbook.add_format({
            'bold': True,
            'text_wrap': True,
            'valign': 'center',
            'font_name': 'Consolas',
            'font_size': 11,
            'border': 1,
            'bottom': 2,
        })
        font_formats = workbook.add_format({'font_name': 'Consolas', 'font_size': 11, })  # 其他

        def write_to_excel_by_dict(scorecard, worksheet, start_row, start_column):
            for featurename, interval_value in scorecard.items():
                if isinstance(interval_value, dict):
                    worksheet.write(start_row, start_column, featurename, font_formats)
                    for interval, value in interval_value.items():
                        woe, score = value[0], value[1]
                        worksheet.write(start_row, start_column + 1, interval, font_formats)
                        worksheet.write(start_row, start_column + 2, woe, font_formats)
                        worksheet.write(start_row, start_column + 3, score, font_formats)
                        start_row += 1
                elif isinstance(interval_value, list):
                    nothing, score = interval_value[0], interval_value[1]
                    worksheet.write(start_row, start_column, featurename, font_formats)
                    worksheet.write(start_row, start_column + 1, nothing, font_formats)
                    worksheet.write(start_row, start_column + 2, nothing, font_formats)
                    worksheet.write(start_row, start_column + 3, score, font_formats)
                    start_row += 1

        # 写入标题
        for col_num, value in enumerate(['特征名称', '特征区间', '特征区间WOE值', '特征区间得分']):
            worksheet.write(0, col_num, value, header_format)

        # 写入数据
        start_row, start_column = 1, 0
        write_to_excel_by_dict(scorecard, worksheet, start_row, start_column)
        worksheet.set_column(0, 1, 20)


def get_object_value_by_train_lr(X_train, y_train, X_test, y_test, object_func='default'):
    """
    计算目标值
    Args:
        X_train:
        y_train:
        X_test:
        y_test:
        object_func:

    Returns:

    """

    lr = LogisticRegression().fit(X_train, y_train)
    y_test_pred = lr.predict_proba(X_test)[:, 1]
    test_ks = get_ks(y_test, y_test_pred)
    test_auc = get_auc(y_test, y_test_pred)

    if object_func == "test_ks":
        return test_ks
    elif object_func == "test_auc":
        return test_auc
    elif object_func == "default":
        return test_ks, test_auc


def get_max_value_in_dic(dic):
    """
    返回字典中最大value的k,v
    Args:
        dic:

    Returns:

    """
    import operator
    key = max(dic.items(), key=operator.itemgetter(1))[0]
    return key, dic[key]


def calculate_features_p_value_vif(df, features, target, just_pvalue=False, need_coef=False, need_model=False):
    """
    计算每个变量的p-value和vif和coef
    Args:
        df:
        features:
        target:
        just_pvalue:
        need_coef:
        need_model:

    Returns:

    """
    import autobmt

    index = 1
    while True:  # 循环的目的是保证入模变量的系数都为整
        model = sm.Logit(df[target], sm.add_constant(df[features])).fit()
        ignore_features = [k for k, v in model.params.to_dict().items() if k != 'const' and v < 0]
        if len(ignore_features) == 0:
            break
        features = [i for i in features if i not in ignore_features]
        index += 1
    df_pvalue = pd.DataFrame(model.pvalues, columns=['pvalue'])
    if just_pvalue:
        return df_pvalue
    df_coef = pd.DataFrame(model.params, columns=['coef'])
    df_vif = pd.Series(index=features)
    for col in features:
        df_vif[col] = pd.Series(autobmt.get_vif(df[list(set(features) - set([col]))], df[col]))
    df_vif = pd.DataFrame(df_vif, columns=['vif'])
    if need_coef:
        df_pvalue_vif_coef = df_coef.merge(df_pvalue, how='left', left_index=True, right_index=True) \
            .merge(df_vif, how='left', left_index=True, right_index=True)
        if need_model:
            return df_pvalue_vif_coef, model, features
        else:
            return df_pvalue_vif_coef, features
    else:
        df_pvalue_vif = df_pvalue.merge(df_vif, how='inner', left_index=True, right_index=True)
        return df_pvalue_vif, features



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-


import os
import sys
module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path)

import numpy as np
import pandas as pd
import pytest

from autobmt.statistics import WOE, calc_iv, _IV, get_vif, calc_var_summary, get_iv_psi, calc_bin_summary

np.random.seed(1)

feature = np.random.rand(500)
target = np.random.randint(2, size=500)
A = np.random.randint(100, size=500)
B = np.random.randint(100, size=500)
mask = np.random.randint(8, size=500)

df = pd.DataFrame({
    'feature': feature,
    'target': target,
    'A': A,
    'B': B,
})


def test_woe():
    value = WOE(0.2, 0.3)
    assert value == pytest.approx(-0.4054651081081643)


def test_iv_priv():
    value = _IV(df['feature'], df['target'])
    assert value == pytest.approx(0.010385942643745403)


def test_iv():
    value = calc_iv(df['feature'], df['target'], n_bins=10, method='dt')
    assert value == pytest.approx(0.23561308290317584)


def test_iv_frame():
    res = calc_iv(df, 'target', n_bins=10, method='chi')
    assert res.loc[0, 'A'] == pytest.approx(0.04517306679551927)


def test_vif():
    vif = get_vif(df)
    assert vif['A'] == 2.9515204239842703


def test_calc_var_summary():
    var_s = calc_var_summary(df, target='target')
    assert list(var_s.loc[4, 'IV'].values) == [0.036723641370124835, 0.0605206894133002, 0.07822121543008585]


def test_get_iv_psi():
    df['type'] = 'train'
    var_iv_psi = get_iv_psi(df, feature_list=['feature', 'A', 'B'], by_col='type', only_psi=False)
    assert list(var_iv_psi.IV) == [0.036723641370124835, 0.0605206894133002, 0.07822121543008587]


dff = pd.DataFrame({
    'feature': feature,
    'target': target,
})


def test_calc_bin_summary():
    result = calc_bin_summary(dff, bin_col='feature', target='target')
    assert result.loc[4, 'ks'] == -0.028036335090277087


# def test_calc_bin_summary_use_step():
#     result = calc_bin_summary(df,bin_col='feature', target='target', method = 'step', clip_q = 0.01)
#     assert result.loc[4, 'ks'] == -0.0422147102645028

def test_calc_bin_summary_for_all_score():
    result = calc_bin_summary(dff, bin_col='feature', target='target', bin=False)
    assert len(result) == 500


def test_calc_bin_summary_return_splits():
    result = calc_bin_summary(dff, bin_col='feature', target='target')
    assert len(result) == 10


def test_calc_bin_summary_use_split_pointers():
    result = calc_bin_summary(dff, bin_col='feature', target='target', bin=[0.2, 0.6])
    assert len(result) == 3


def test_calc_bin_summary_with_lift():
    result = calc_bin_summary(dff, bin_col='feature', target='target', is_sort=False)
    assert result.loc[3, 'lift'] == 1.0038610038610039


def test_calc_bin_summary_with_cum_lift():
    result = calc_bin_summary(dff, bin_col='feature', target='target', is_sort=False)
    assert result.loc[3, 'cum_lift'] == 1.003861003861004



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: plot.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-10-18
'''
import os

import matplotlib.pyplot as plt
import numpy as np
import openpyxl
import pandas as pd
import six
from matplotlib import gridspec
from openpyxl.drawing.image import Image

from .statistics import calc_var_summary


def plot_var_bin_summary(frame, cols, target='target', by='type', file_path=None, sheet_name='plot_var_bin_summary',
                         need_bin=False, **kwargs):
    """
    画变量分箱图
    Args:
        frame (DataFrame):
        cols (str|list): 需要画图的变量
        target (str): 目标变量
        by (str): 分开统计的列名
        file_path (str): 保存路径
        sheet_name (str):保存excel的sheet名称
        need_bin (bool):是否要进行分箱

    Returns:

    """
    # frame = {'train':tr_df,'test':te_df,'oot':oot_df}

    if isinstance(cols, str):
        cols = [cols]

    if isinstance(frame, pd.DataFrame):
        summary_df_dict = {}
        if by in frame:
            for name, df in frame.groupby(by):
                summary_df_dict[name] = calc_var_summary(df, include_cols=cols, target=target, need_bin=need_bin,
                                                         **kwargs)
        else:
            summary_df_dict['all'] = calc_var_summary(frame, include_cols=cols, target=target, need_bin=need_bin,
                                                      **kwargs)
    else:
        summary_df_dict = frame

    save_jpg_path = None
    if file_path is not None:
        if '.xlsx' in file_path:
            fp_sp = file_path.split('.xlsx')
            if '' == fp_sp[1]:
                save_jpg_path = fp_sp[0] + '_var_jpg'
                os.makedirs(save_jpg_path, exist_ok=True)

            else:
                save_jpg_path = file_path + '/var_jpg'
                os.makedirs(save_jpg_path, exist_ok=True)
                file_path = file_path + '/plot_var_bin_summary.xlsx'
        else:
            save_jpg_path = file_path + '/var_jpg'
            os.makedirs(save_jpg_path, exist_ok=True)
            file_path = file_path + '/plot_var_bin_summary.xlsx'

    for col in cols:
        # 做图
        # plt.figure(figsize=(15, 4))
        # gs = gridspec.GridSpec(1, 2)
        # gs = gridspec.GridSpec(1, 3)

        gs_num = len(summary_df_dict)

        if gs_num == 1:
            plt.figure(figsize=(gs_num * 5, 4), dpi=400)
        else:
            plt.figure(figsize=(gs_num * 5, 4))

        gs = gridspec.GridSpec(1, gs_num)

        for i, k in enumerate(summary_df_dict):

            df = summary_df_dict[k]

            if i == 0:
                ax1 = plt.subplot(gs[0, 0])
            else:
                ax1 = plt.subplot(gs[0, i], sharey=ax1)

            df = df[df['var_name'] == col]

            x_PctTotal = np.arange(0, len(df), 1)
            y_PctTotal = df['total_pct']

            ax1.bar(x_PctTotal, y_PctTotal, color='lightskyblue', alpha=0.4, width=0.5, label='PctTotal')

            for x, y in zip(x_PctTotal, y_PctTotal):
                ax1.text(x + 0.05, y + 0.01, str(round(y * 100, 2)) + '%', ha='center', va='bottom', fontsize=12)
            ax1.set_ylabel('total_pct', fontsize=12)

            ax1.set_ylim([0, max(y_PctTotal) + 0.3])
            ax2 = ax1.twinx()
            ax2.plot(x_PctTotal, df['positive_rate'], '-ro', color='red')

            for x, y in zip(x_PctTotal, df['positive_rate']):
                ax2.text(x + 0.05, y, str(round(y * 100, 2)) + '%', ha='center', va='bottom', fontsize=12, color='r')
            ax2.set_ylabel('positive_rate', fontsize=12)
            ax2.set_ylim([0, max(df['positive_rate']) + 0.01])

            # title_map = {0: 'train', 1: 'test', 2: 'oot'}
            # plt.title('{}:{}'.format(title_map[i], col))
            plt.title('{}:{}\nIV: {:.5f}'.format(k, col, df['IV'].iloc[0]))
            # plt.title('train:' + col)
            # 将数据详情表添加
            tmp_df = df[['range', 'woe', 'iv', 'total', 'positive_rate']]
            round_cols = ['woe', 'iv', 'positive_rate']
            tmp_df[round_cols] = tmp_df[round_cols].applymap(lambda v: round(v, 4) if pd.notnull(v) else '')
            mpl_table = plt.table(cellText=tmp_df.values, colLabels=tmp_df.columns,
                                  colWidths=[0.15, 0.1, 0.1, 0.1, 0.15],
                                  loc=1)  # loc='top'将详情放到顶部

            mpl_table.auto_set_font_size(False)
            mpl_table.set_fontsize(5)

            header_color = '#40466e'
            row_colors = ['#f1f1f2', 'w']
            header_columns = 0

            for k, cell in six.iteritems(mpl_table._cells):
                cell.set_edgecolor('w')
                if k[0] == 0 or k[1] < header_columns:
                    cell.set_text_props(weight='bold', color='w')
                    cell.set_facecolor(header_color)
                else:
                    cell.set_facecolor(row_colors[k[0] % len(row_colors)])

            # ax1.text(*(ax1.get_xlim()[0], ax1.get_ylim()[1]), 'IV: {:.5f}'.format(df['iv'].sum()), fontsize='x-large')

        plt.tight_layout()

        if save_jpg_path is not None:  # 判断是否需要保存
            plt.savefig(save_jpg_path + '/{}.jpg'.format(col), dpi=300)  # dpi控制清晰度
        # plt.show()

    # if save_jpg_path is not None:
    #     workbook = xlsxwriter.Workbook(file_path)
    #     worksheet = workbook.add_worksheet(sheet_name)
    #
    #     for i, jpg_name in enumerate(cols):
    #         worksheet.insert_image('A{}'.format(i * 29 + 3), save_jpg_path + '/{}.jpg'.format(jpg_name),
    #                                {'x_scale': 1.5, 'y_scale': 1.5})
    #
    #     workbook.close()
    if save_jpg_path is not None:
        try:
            wb = openpyxl.load_workbook(file_path)
        except:
            wb = openpyxl.Workbook()
        sh = wb.create_sheet(sheet_name)
        for i, jpg_name in enumerate(cols):
            img = Image(save_jpg_path + '/{}.jpg'.format(jpg_name))
            newsize = (1900, 500)
            img.width, img.height = newsize  # 设置图片的宽和高
            sh.add_image(img, 'A{}'.format(i * 29 + 3))
        wb.save(file_path)



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: scorecard.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-05-20
'''

import re
from copy import deepcopy

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.linear_model import LogisticRegression

from .transformer import FeatureBin, WoeTransformer
from .utils import to_ndarray, save_json, load_json

RE_NUM = r'-?\d+(.\d+)?'
RE_SEP = r'[~-]'
RE_BEGIN = r'(-inf|{num})'.format(num=RE_NUM)
RE_END = r'(inf|{num})'.format(num=RE_NUM)
RE_RANGE = r'\[{begin}\s*{sep}\s*{end}\)'.format(
    begin=RE_BEGIN,
    end=RE_END,
    sep=RE_SEP,
)

NUMBER_EXP = re.compile(RE_RANGE)

NUMBER_EMPTY = -9999999
NUMBER_INF = 1e10
FACTOR_EMPTY = 'MISSING'
FACTOR_UNKNOWN = 'UNKNOWN'
ELSE_GROUP = 'else'


class ScoreCard(BaseEstimator):
    def __init__(self, pdo=50, rate=2, odds=15, base_score=600,
                 card={}, combiner={}, transer=None, **kwargs):
        """

        Args:
            pdo (int):point double odds;;;当odds增加一倍，评分增加的分数
            rate (int):
            odds (int): odds at base point;;;基准分值对应的odds
            base_score (int): base point;;;基准分数
            card (dict): 评分卡
            combiner (autobmt.FeatureBin): 分箱规则
            transer (autobmt.WoeTransformer): 变量分箱对应的woe值
            **kwargs:
        """
        self.pdo = pdo  # point double odds;;;当odds增加一倍，评分增加的分数
        self.rate = rate  #
        self.odds = odds  # odds at base point;;;基准分值对应的odds
        self.base_score = base_score  # base point;;;基准分数

        self.factor = pdo / np.log(rate)  # 大B;;;B=72.13475204444818
        self.offset = base_score - (pdo / np.log(rate)) * np.log(odds)  # 大A;;;A=404.65547021957406

        self._combiner = combiner
        self.transer = transer
        self.model = LogisticRegression(**kwargs)

        self._feature_names = None

        self.card = card
        if card:
            self.load(card)

    def __len__(self):
        return len(self.card.keys())

    def __contains__(self, key):
        return key in self.card

    def __getitem__(self, key):
        return self.card[key]

    def __setitem__(self, key, value):
        self.card[key] = value

    def __iter__(self):
        return iter(self.card)

    @property
    def coef_(self):
        """ 逻辑回归模型系数
        """
        return self.model.coef_[0]

    @property
    def intercept_(self):
        """ 逻辑回归模型截距
        """
        return self.model.intercept_[0]

    @property
    def n_features_(self):
        """ 变量个数
        """
        return (self.coef_ != 0).sum()

    @property
    def features_(self):
        """ 变量列表
        """
        if not self._feature_names:
            self._feature_names = list(self.card.keys())

        return self._feature_names

    @property
    def combiner(self):
        if not self._combiner:
            # 如果不存在，则生成新的分箱器
            rules = {}
            for key in self.card:
                rules[key] = self.card[key]['bins']

                self._combiner = FeatureBin().manual_bin(rules)

        return self._combiner

    def fit(self, X, y):
        """
        Args:
            X (2D DataFrame): 变量
            Y (array-like): 目标变量列表
        """
        self._feature_names = X.columns.tolist()

        for f in self.features_:
            if f not in self.transer:
                raise Exception('column \'{f}\' is not in transer'.format(f=f))

        self.model.fit(X, y)
        self.card = self._generate_rules()

        # keep sub_score-median of each feature, as `base_effect` for reason-calculation
        sub_score = self.woe_to_score(X)
        # self.base_effect = pd.Series(
        #     np.median(sub_score, axis=0),
        #     index=self.features_
        # )

        return self

    def predict(self, X, **kwargs):
        """预测分数
        Args:
            X (2D-DataFrame|dict): 需要去预测的变量
            return_sub (Bool): 是否需要返回特征中每个箱子的得分
            default (str|number): 未知特征的默认分数，' min '(默认)，' max '

        Returns:
            array-like: 预测的分数
            DataFrame|dict: 每个特征对应的分数
        """
        if isinstance(X, pd.DataFrame):
            X = X[self.features_]

        bins = self.combiner.transform(X)
        res = self.bin_to_score(bins, **kwargs)
        return res

    def bin_to_score(self, bins, return_sub=False, default='min'):
        """
        通过分箱值直接算分
        Args:
            bins (2D-DataFrame|dict): 使用分箱值替换后的变量
            return_sub (bool): 是否需要返回特征中每个箱子的得分
            default (bool): 未知特征的默认分数，' min '(默认)，' max '

        Returns:

        """
        score = 0
        res = bins.copy()
        for col, rule in self.card.items():
            s_map = rule['scores']
            b = bins[col]

            # set default value for empty group
            default_value = default
            if default == 'min':
                default_value = np.min(s_map)
            elif default == 'max':
                default_value = np.max(s_map)
            elif isinstance(default, str):
                raise ValueError(f'default `{default}` is not valid, only support `min`, `max` or number')

            # append default value to the end of score map
            s_map = np.append(s_map, default_value)

            # # set default group to min score
            # if np.isscalar(b):
            #     b = np.argmin(s_map) if b == self.EMPTY_BIN else b
            # else:
            #     b[b == self.EMPTY_BIN] = np.argmin(s_map)

            # replace score
            res[col] = s_map[b]
            score += s_map[b]

        if return_sub:
            return score, res
        else:
            return score

    def predict_proba(self, X):
        """
        预测概率
        Args:
            X (2D array-like): 需要去预测的变量

        Returns:
            2d array: 预测的概率值（包括正样本和负样本的概率值）
        """
        proba = self.score_to_proba(self.predict(X))
        return np.stack((1 - proba, proba), axis=1)

    def _generate_rules(self):
        if not self._check_rules(self.combiner, self.transer):
            raise Exception('generate failed')

        rules = {}

        for idx, key in enumerate(self.features_):
            weight = self.coef_[idx]

            if weight == 0:
                continue

            # woe = self.transer[key]['woe']
            woe = list(self.transer[key].values())

            rules[key] = {
                'bins': self.combiner[key],
                'woes': woe,
                'weight': weight,
                'scores': self.woe_to_score(woe, weight=weight),
            }

        return rules

    def _check_rules(self, combiner, transer):
        for col in self.features_:
            if col not in combiner:
                raise Exception('column \'{col}\' is not in combiner'.format(col=col))

            if col not in transer:
                raise Exception('column \'{col}\' is not in transer'.format(col=col))

            l_c = len(combiner.export(bin_format=False)[col])
            # l_t = len(transer[col]['woe'])
            l_t = len(transer[col].values())

            if l_c == 0:
                continue

            if np.issubdtype(combiner[col].dtype, np.number):
                if l_c != l_t - 1:
                    raise Exception(
                        'column \'{col}\' is not matched, assert {l_t} bins but given {l_c}'.format(col=col, l_t=l_t,
                                                                                                    l_c=l_c + 1))
            else:
                if l_c != l_t:
                    raise Exception(
                        'column \'{col}\' is not matched, assert {l_t} bins but given {l_c}'.format(col=col, l_t=l_t,
                                                                                                    l_c=l_c))

        return True

    def proba_to_score(self, prob):
        """概率转分

        odds = (1 - prob) / prob    #good:bad
        score = factor * log(odds) + offset

        odds = prob / (1 - prob)    #bad:good
        score = offset - factor * log(odds)

        log(odds) = intercept+woe1*coef1+woe2*coef2+woe3*coef3
        """
        # return self.factor * np.log((1 - prob) / prob) + self.offset
        return self.offset - self.factor * np.log(prob / (1 - prob))

    def woe_sum_to_score(self, woe, weight=None):
        """通过woe计算分

        odds = (1 - prob) / prob    #good:bad
        score = factor * log(odds) + offset

        odds = prob / (1 - prob)    #bad:good
        score = offset - factor * log(odds)

        log(odds) = intercept+woe1*coef1+woe2*coef2+woe3*coef3
        """
        woe = to_ndarray(woe)

        if weight is None:
            weight = self.coef_

        z_cols = weight * woe
        z = self.intercept_ + np.sum(z_cols, axis=1)

        return self.offset - self.factor * z

    def score_to_proba(self, score):
        """分转概率

        Returns:
            array-like|float: 正样本的概率【即：1的概率】
        """
        return 1 / (1 + np.exp((score - self.offset) / self.factor))

    def woe_to_score(self, woe, weight=None):
        """通过woe计算分
        score = A - Blog(odds) = A - B( β0 + β1x1 + … βnxn) = (A - Bβ0) - Bβ1 x1 - … Bβn xn = sum((A - Bβ0)/n - Bβ1 x1 - … Bβn xn)
        """
        woe = to_ndarray(woe)

        if weight is None:
            weight = self.coef_

        b = (self.offset - self.factor * self.intercept_) / self.n_features_  # (A - Bβ0)/n
        s = -self.factor * weight * woe  # -B*βn*xn

        # drop score whose weight is 0
        mask = 1
        if isinstance(weight, np.ndarray):
            mask = (weight != 0).astype(int)

        return (s + b) * mask

    def export(self, to_dataframe=False, to_json=None, to_csv=None, decimal=2):
        """生成一个评分卡对象

        Args:
            to_dataframe (bool): 生成的评分卡是1个DataFrame
            to_json (str|IOBase): 生成的评分卡写出json文件
            to_csv (filepath|IOBase): 生成的评分卡写出csv文件

        Returns:
            dict: 评分卡
        """
        card = dict()
        combiner = self.combiner.export(bin_format=True, index=False)

        for col, rule in self.card.items():
            s_map = rule['scores']
            bins = combiner[col]
            woe_map = np.zeros(len(bins))
            if 'woes' in rule:
                woe_map = rule['woes']
            weight = np.nan
            if 'weight' in rule:
                weight = rule['weight']
            card[col] = dict()

            for i in range(len(bins)):
                # card[col][bins[i]] = round(s_map[i], decimal)
                card[col][bins[i]] = [round(s_map[i], decimal), woe_map[i], weight]

        if to_json is not None:
            save_json(card, to_json)

        if to_dataframe or to_csv is not None:
            rows = list()
            for feature in card:
                for value, score in card[feature].items():
                    rows.append({
                        'feature': feature,
                        'value': value,
                        'score': score[0],
                        'woe': score[1],
                        'weight': score[2],
                    })

            card = pd.DataFrame(rows)

        if to_csv is not None:
            return card.to_csv(to_csv)

        return card

    def _is_numeric(self, bins):
        m = NUMBER_EXP.match(bins[0])

        return m is not None

    def _numeric_parser(self, bins):
        l = list()

        for item in bins:

            # if re.compile('{}.nan'.format(RE_NUM)).match(item):
            if item == 'nan':
                l.append(np.nan)
                continue

            m = NUMBER_EXP.match(item)
            split = m.group(3)

            if split == 'inf':
                # split = np.inf
                continue

            split = float(split)

            l.append(split)

        return np.array(l)

    def parse_bins(self, bins):
        """解析格式化的分箱值
        """
        if self._is_numeric(bins):
            return self._numeric_parser(bins)

        l = list()

        for item in bins:
            if item == ELSE_GROUP:
                l.append(item)
            else:
                l.append(item.split(','))

        return np.array(l, dtype=object)

    def _parse_rule(self, rule):
        bins = self.parse_bins(list(rule.keys()))
        v = list(rule.values())
        if isinstance(v[0], list):
            scores = np.array([i[0] for i in v])
        else:
            scores = np.array(v)

        return {
            'bins': bins,
            'scores': scores,
        }

    def load(self, card=None):
        """
        加载评分卡
        Args:
            card: 从dict或json文件加载评分卡

        Returns:

        """
        card = deepcopy(card)
        if not isinstance(card, dict):
            card = load_json(card)

        for feature in card:
            card[feature] = self._parse_rule(card[feature])

        self.card = card
        self._combiner = {}

        return self



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: utils.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-05-20
'''

import gc
import json
import math
import pickle
import re
import sys
import warnings

import numpy as np
import pandas as pd
import statsmodels
import statsmodels.api as sm
# from catboost import CatBoostRegressor
# from lightgbm import LGBMRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import roc_curve, accuracy_score, r2_score, recall_score, precision_score, f1_score
from sklearn.model_selection import train_test_split
#from xgboost.sklearn import XGBRegressor
from functools import wraps

import autobmt

warnings.filterwarnings('ignore')

from .logger_utils import Logger

log = Logger(level="info", name=__name__).logger

FILLNA = -999
CONTINUOUS_NUM = 10


def support_dataframe(require_target=True):
    """用于支持dataframe的装饰器
    """

    def sup_df(fn):
        @wraps(fn)
        def func(frame, *args, **kwargs):
            if not isinstance(frame, pd.DataFrame):
                return fn(frame, *args, **kwargs)

            frame = frame.copy()
            if require_target and isinstance(args[0], str):
                target = frame.pop(args[0])
                args = (target,) + args[1:]
            elif 'target' in kwargs and isinstance(kwargs['target'], str):
                kwargs['target'] = frame.pop(kwargs['target'])

            if 'return_bin' not in kwargs:
                kwargs['return_bin'] = True

            res = dict()
            for col in frame:
                r = fn(frame[col], *args, **kwargs)

                if not isinstance(r, np.ndarray):
                    if isinstance(r, tuple):
                        r = r[1]
                    else:
                        r = [r]
                res[col] = r
            return pd.DataFrame(res)

        return func

    return sup_df


def get_splitted_data(df_selected, target, selected_features):
    X = {}
    y = {}

    X['all'] = df_selected[selected_features]
    y['all'] = df_selected[target]

    for name, df in df_selected.groupby('type'):
        X[name] = df[selected_features]
        y[name] = df[target]

    if not X.__contains__('oot'):
        X['oot'] = None
        y['oot'] = None

    return X['all'], y['all'], X['train'], y['train'], X['test'], y['test'], X['oot'], y['oot']


# def to_score(x, A=404.65547022, B=72.1347520444):
#     result = round(A - B * math.log(x / (1 - x)))
#
#     if result < 0:
#         result = 0
#     if result > 1200:
#         result = 1200
#     result = 1200 - result
#     return result

def to_score(x, A=404.65547022, B=72.1347520444):
    """
    概率值转分
    Args:
        x (float): 模型预测的概率值
        pdo=50;rate=2;odds=15;base_score=600
        A (float): 评分卡offset；；；offset = base_score - (pdo / np.log(rate)) * np.log(odds)
        B (float): 评分卡factor；；；factor = pdo / np.log(rate)

    Returns:
        score (float): 标准评分
    """
    result = round(A - B * math.log(x / (1 - x)))

    # if result < 300:
    #     result = 300
    # if result > 900:
    #     result = 900
    return result


def score2p(x, A=404.65547022, B=72.1347520444):
    """
    分转概率
    Args:
        x (float): 标准评分
        pdo=50;rate=2;odds=15;base_score=600
        A (float): 评分卡offset；；；offset = base_score - (pdo / np.log(rate)) * np.log(odds)
        B (float): 评分卡factor；；；factor = pdo / np.log(rate)

    Returns:
        p (float): 概率值
    """
    return 1 / (1 + np.exp((x - A) / B))


def train_test_split_(df_src, target='target', test_size=0.3):
    """
    样本切分函数.先按target分类，每类单独切成train/test，再按train/test合并，
    使得train/test的badrate能高度一致
    Args:
        df_src:
        target:
        test_size:

    Returns:

    """

    l = [[], [], [], []]
    for target_value, X in df_src.groupby(target):

        X[target] = target_value

        row = train_test_split(X.drop(labels=target, axis=1), X[target], test_size=test_size, random_state=1024)

        for i in range(0, 4):
            l[i].append(row[i])

    list_df = []
    for i in range(0, 4):
        list_df.append(pd.concat(l[i]))

    return tuple(list_df)


def split_data_type(df, key_col='id', target='target', apply_time='apply_time', test_size=0.3):
    if df[target].isin([0, 1]).all():
        log.info('样本y值在0，1')
    else:
        log.info('\033[0;31m样本y值不在0，1之间，请检查！！！\033[0m')

    assert df[target].isin([0, 1]).all()

    log.info('样本情况：', df.shape)
    df.drop_duplicates(subset=key_col, inplace=True)
    log.info('分布情况：', df.groupby(target)[key_col].count().sort_index())
    log.info('样本drop_duplicates情况：', df.shape)

    # ---------查看各月badrate---------------------
    df['apply_month'] = df[apply_time].map(lambda s: s[:7])
    log.info(df.groupby('apply_month').describe()[target])
    del df['apply_month']

    # ---------样本划分----------------------------
    ##需要oot
    # df_selected = df_id #can filter records here
    # # df_oot = df_selected[df_selected['apply_time']>= '2019-04-01']
    # # X_train = df_selected[df_selected['apply_time']<= '2019-01-31']
    # # X_test = df_selected[(df_selected['apply_time']> '2019-01-31') & (df_selected['apply_time']< '2019-04-01')]

    # df_oot = df_selected[df_selected['apply_time']>= '2019-03-01']
    # X_train = df_selected[df_selected['apply_time']<= '2018-12-31']
    # X_test = df_selected[(df_selected['apply_time']> '2018-12-31') & (df_selected['apply_time']< '2019-03-01')]

    # #X_train, X_test, y_train, y_test = geo_train_test_split(df_not_oot,label=label)

    # df_id.loc[df_oot.index,'type'] = 'oot'
    ##需要oot

    # 不需要oot的时候运行下面这一行代码
    X_train, X_test, y_train, y_test = train_test_split_(df, target=target, test_size=test_size)
    # X_train, X_test, y_train, y_test = train_test_split(df_id.drop(columns=target), df_id[target], test_size=test_size,
    #                                                     random_state=123)
    # 不需要oot的时候运行下面这一行代码

    df.loc[X_train.index, 'type'] = 'train'
    df.loc[X_test.index, 'type'] = 'test'

    log.info(df.groupby('type').describe()[target])

    # ----------输出---------------------------------
    # df_id.to_csv(data_dir + '{}_split.csv'.format(client_batch), index=False)
    return df


def select_features_dtypes(df, exclude=None):
    """
    根据数据集，筛选出数据类型
    Args:
        df: 数据集
        exclude: 排除不需要参与筛选的列

    Returns:

    """
    if exclude is not None:
        df = df.drop(columns=exclude)
    # 筛选出数值类型列
    numeric_list = df.select_dtypes([np.number]).columns.tolist()

    no_numeric_df = df.select_dtypes(include=['object'])
    # 将object类型的列尝试转成时间类型
    dates_objs_df = no_numeric_df.apply(pd.to_datetime, errors='ignore')
    # 筛选出字符类型列
    objs_list = dates_objs_df.select_dtypes(include=['object']).columns.tolist()
    # 筛选出时间类型列
    # dates_df = list(set(dates_objs_df.columns) - set(objs_df.columns))
    date_list = dates_objs_df.select_dtypes(include=['datetime64']).columns.tolist()

    assert len(numeric_list) + len(objs_list) + len(date_list) == df.shape[1]

    return numeric_list, objs_list, date_list


def filter_miss(df, miss_threshold=0.9):
    """

    Args:
        df (DataFrame): 用于训练模型的数据集
        miss_threshold: 缺失率大于等于该阈值的变量剔除

    Returns:

    """
    names_list = []
    for name, series in df.items():
        n = series.isnull().sum()
        miss_q = n / series.size
        if miss_q < miss_threshold:
            names_list.append(name)
    return names_list


###################
def step_evaluate_models(df, features, target, stepname="", is_turner=False):
    """
    用lr,xgb,评估train/test/oot数据集的auc,ks
    Args:
        df: 数据集，包含y,type,features
        features: 入模特征
        target: 目标值
        stepname: 标识是在哪一步进行评估的
        is_turner: 是否需要进行调参

    Returns:

    """
    X_train = df[df['type'] == 'train'][features]
    y_train = df[df['type'] == 'train'][target]
    data = df[['type', target]]

    # xgb默认参数
    xgb_params = {"base_score": 0.5, "booster": "gbtree", "colsample_bylevel": 1, "colsample_bytree": 0.8, "gamma": 3,
                  "learning_rate": 0.1, "max_delta_step": 0, "max_depth": 6, "min_child_weight": 50,
                  "n_estimators": 200, "n_jobs": -1, "objective": "binary:logistic", "random_state": 0,
                  "reg_alpha": 5, "reg_lambda": 5, "scale_pos_weight": 1,
                  "subsample": 0.8}
    lightgbm_params = {'boosting_type': 'gbdt', 'num_threads': 20,
                       'min_child_weight': 50, 'max_depth': 6,
                       'colsample_bytree': 0.8, 'subsample': 0.8,
                       'num_iterations': 200, 'learning_rate': 0.1, 'verbose': -1
                       }
    rf_params = {'max_depth': 6,
                 'n_estimators': 200,
                 'min_samples_leaf': 60, 'n_jobs': -1, 'min_samples_split': 60,
                 'verbose': 0
                 }
    catboost_params = {'depth': 6, 'l2_leaf_reg': 3,
                       'n_estimators': 200, 'learning_rate': 0.1,
                       'subsample': 0.8
                       }
    if is_turner:
        # 需要调参走这个逻辑
        # models = {
        #     "lr": LogisticRegression().fit(X_train, y_train),
        #     "rf": rf_turner(X_train, y_train),
        #     "xgb": xgb_turner(X_train, y_train, X_test, y_test)[1],
        #     # "lightgbm": lightgbm_turner(X_train, y_train, X_test, y_test)[1],
        #     # "catboost": catboost_turner(X_train, y_train, X_test, y_test)[1]
        # }

        pass
    else:
        # 使用默认参数进行训练评估
        models = {
            "lr": LogisticRegression().fit(X_train, y_train),
            "rf": RandomForestRegressor(**rf_params).fit(X_train, y_train),
            #"xgb": XGBRegressor(**xgb_params).fit(X_train, y_train),
            # "lightgbm": LGBMRegressor(**lightgbm_params).fit(X_train, y_train),
            # "catboost": CatBoostRegressor(**catboost_params, verbose=False).fit(X_train, y_train)
        }

    result = []
    for name, model in models.items():
        # model = model.fit(X_train, y_train)
        if isinstance(model, LogisticRegression):
            data['prob'] = model.predict_proba(df[features])[:, 1]
        elif isinstance(model, RandomForestRegressor):
            data['prob'] = model.predict(df[features])
        # elif isinstance(model, XGBRegressor):
        #     data['prob'] = model.predict(df[features])
        # elif isinstance(model, LGBMRegressor):
        #     data['prob'] = model.predict(df[features])
        # elif isinstance(model, CatBoostRegressor):
        #     data['prob'] = model.predict(df[features])

        df_splitted_type_auc_ks = data.groupby('type').apply(
            lambda df_: pd.Series({'{}_auc'.format(name): autobmt.get_auc(df_[target], df_['prob']),
                                   '{}_ks'.format(name): autobmt.get_ks(df_[target], df_['prob'])}))
        result.append(df_splitted_type_auc_ks)

    evaluate_df = pd.concat(result, axis=1)
    one_step_df = merge_rows_one_row_df(evaluate_df)
    one_step_df['feature_num'] = len(features)  # 加上特征个数
    one_step_df['stepname'] = stepname  # 用于标识是哪一步计算的auc和ks
    xgb_evaluate_log = str(one_step_df.filter(like='xgb').applymap(lambda x: round(x, 4)).iloc[0].to_dict())
    del_df(data)
    del_df(X_train)
    del_df(y_train)
    return one_step_df, xgb_evaluate_log


def model_predict_evaluate(model, df, features, target, A=404.65547022, B=72.1347520444, exclude_cols=None,
                           is_return_var=False):
    """
    1. 利用已经训练好的model对象，对数据集进行预测
    2. 数据集中需要包含有一列type，标识train/test/oot
    3. 返回各个数据集的auc,ks
    4. 返回type,真实y值，预测的概率值，预测的标准分(供后续输入到模型报告中生成外部报告)
    LogisticRegression().fit()
    model = sm.Logit(train_data[target], sm.add_constant(train_data[features])).fit()
    XGBRegressor().fit()
    Args:
        model: 已经训练好的模型对象，支持lr,xgb
        df: 用于训练模型的数据集
        features: 入模变量
        target: 目标值
        A: 评分卡大A
        B: 评分卡大B
        exclude_cols: 样本中的原始字段（非x，y）

    Returns:

    """
    assert 'type' in df.columns.tolist()
    data = df.copy()
    if isinstance(model, LogisticRegression):
        data['p'] = model.pa(data[features])[:, 1]
    elif isinstance(model, statsmodels.discrete.discrete_model.BinaryResultsWrapper):
        log.info("mode is : statsmodels.discrete.discrete_model.BinaryResultsWrapper")
        data['p'] = model.predict(sm.add_constant(data[features]))
    # elif isinstance(model, XGBRegressor):
    #     data['p'] = model.predict(data[features])
    # elif isinstance(model, LGBMRegressor):
    #     data['p'] = model.predict(data[features])
    # elif isinstance(model, CatBoostRegressor):
    #     data['p'] = model.predict(data[features])
    elif isinstance(model, RandomForestRegressor):
        data['p'] = model.predict(data[features])
    data['score'] = data['p'].map(lambda x: to_score(x, A, B))
    evaluate_df = data.groupby('type').apply(
        lambda df: pd.Series({'auc': autobmt.get_auc(df[target], df['p']),
                              'ks': autobmt.get_ks(np.array(df[target]), df['p'])}))

    if is_return_var:
        return evaluate_df, data[exclude_cols + ['p', 'score'] + features]
    else:
        return evaluate_df, data[exclude_cols + ['p', 'score']]


def del_df(df):
    """
    清空一个dataframe
    Args:
        df (DataFrame): 用于训练模型的数据集

    Returns:

    """
    # df.drop(df.index, inplace=True)
    del df
    gc.collect()


def merge_rows_one_row_df(df, name="", stepname=None):
    """将一个多行的dataframe合并成只有一行的dataframe"""
    tmp_arr = []
    for i in range(df.shape[0]):
        tmp = df.iloc[i, :].add_prefix("{}_{}".format(df.index[i], name))
        tmp_arr.append(tmp)

    result_df = pd.DataFrame(pd.concat(tmp_arr, axis=0)).T
    if stepname is not None:  # 合并成一行后,增加一列标识，用于和别的评估进行区分
        result_df['stepname'] = stepname
    return result_df


def get_max_corr_feature(df, features):
    """返回每个变量与其相关性最高的变量以及相关性"""
    corr_df = df.loc[:, features].corr()
    corr_value_series = corr_df.apply(lambda x: x.nlargest(2)[1]).rename("corr_value")
    corr_name_series = corr_df.apply(lambda x: x.nlargest(2).index[1]).rename("corr_name")

    max_corr_df = pd.concat([corr_name_series, corr_value_series], axis=1)
    return max_corr_df


def dump_to_pkl(contents, path):
    pickle.dump(contents, open(path, "wb"))


def load_from_pkl(path):
    return pickle.load(open(path, 'rb'))


def read_sql_string_from_file(path):
    with open(path, 'r', encoding='utf-8') as fb:
        sql = fb.read()
        return sql


def fea_woe_dict_format(fea_woe_dict, splits_dict):
    for j in fea_woe_dict:
        range_format = {int(re.match(r"^(\d+)\.", i).group(1)): i for i in splits_dict[j]}
        fea_woe_dict[j] = {range_format[k]: v for k, v in fea_woe_dict[j].items()}

    return fea_woe_dict


###################

def save_json(res_dict, file, indent=4):
    """
    保存成json文件
    Args:
        res_dict (dict): 需要保存的内容
        file (str|IOBase): 保存后的文件
        indent (int): json文件格式化缩进

    Returns:

    """
    if isinstance(file, str):
        file = open(file, 'w')

    with file as f:
        json.dump(res_dict, f, ensure_ascii = False, indent = indent)


def load_json(file):
    """
    读取json文件
    """
    if isinstance(file, str):
        file = open(file, 'r')

    with file as f:
        res_dict = json.load(f)

    return res_dict


def is_continuous(series):
    series = to_ndarray(series)
    if not np.issubdtype(series.dtype, np.number):
        return False

    n = len(np.unique(series))
    return n > CONTINUOUS_NUM or n / series.size > 0.5
    # return n / series.size > 0.5


def to_ndarray(s, dtype=None):
    """
    """
    if isinstance(s, np.ndarray):
        arr = np.copy(s)
    elif isinstance(s, pd.core.base.PandasObject):
        arr = np.copy(s.values)
    else:
        arr = np.array(s)

    if dtype is not None:
        arr = arr.astype(dtype)
    # covert object type to str
    elif arr.dtype.type is np.object_:
        arr = arr.astype(np.str)

    return arr


def fillna(feature, fillna_va=FILLNA):
    # 复制array 或者 将pandas.core.series.Series变成array
    copy_fea = np.copy(feature)

    mask = pd.isna(copy_fea)

    copy_fea[mask] = fillna_va

    return copy_fea


def split_empty(feature, y=None, return_mask=True):
    copy_fea = np.copy(feature)
    mask = pd.isna(copy_fea)

    copy_fea = copy_fea[~mask]
    if y is not None:
        copy_y = np.copy(y)
        copy_y = copy_y[~mask]
        return copy_fea, copy_y, mask
    return copy_fea, mask


def split_points_to_bin(feature, splits):
    """split points to bin feature
    """
    # log.info("def split_points_to_bin(feature, splits):")
    # log.info(splits)
    feature = fillna(feature)
    return np.digitize(feature, splits)


def np_count(arr, value, default=None):
    c = (arr == value).sum()

    if default is not None and c == 0:
        return default

    return c


def unpack_tuple(x):
    if len(x) == 1:
        return x[0]
    else:
        return x


def split_target(frame, target):
    """
    """
    if isinstance(target, str):
        f = frame.drop(columns=target)
        t = frame[target]
    else:
        f = frame.copy(deep=False)
        t = target

    return f, t


##################g
# corr
def get_corr(df):
    return df.corr()


# accuracy
def get_accuracy(y_true, y_pred):
    return accuracy_score(y_true, y_pred)


# precision
def get_precision(y_true, y_pred):
    if len(np.unique(y_true)) == 2:
        return precision_score(y_true, y_pred)
    else:
        return precision_score(y_true, y_pred, average='macro')


# recall
def get_recall(y_true, y_pred):
    if len(np.unique(y_true)) == 2:
        return recall_score(y_true, y_pred)
    else:
        return precision_score(y_true, y_pred, average='macro')


# f1
def get_f1(y_true, y_pred):
    if len(np.unique(y_true)) == 2:
        return f1_score(y_true, y_pred)
    else:
        return f1_score(y_true, y_pred, average='macro')


# r2
def r2(preds, target):
    return r2_score(target, preds)


def get_best_threshold(y_true, y_pred):
    fpr, tpr, threshold = roc_curve(y_true, y_pred)
    ks = list(tpr - fpr)
    thresh = threshold[ks.index(max(ks))]
    return thresh


def get_bad_rate(df):
    return df.sum() / df.count()


def sigmoid(z):
    return 1 / (1 + np.exp(-z))


##################g


##################考虑用太极实现
def t_cols_sum_axis_1_np(arr):
    res = np.zeros(arr.shape[0], dtype=float)

    for i in range(arr.shape[0]):
        for j in range(arr.shape[1]):
            res[i] += arr[i, j]

    return res


def t_cols_sum_axis_0_np(arr):
    res = np.zeros(arr.shape[1], dtype=float)

    for i in range(arr.shape[0]):
        for j in range(arr.shape[1]):
            res[j] += arr[i, j]

    return res


def t_min_np(arr):
    res = np.inf
    for i in range(arr.shape[0]):
        if res > arr[i]:
            res = arr[i]

    return res


def t_sum_np(arr):
    res = 0
    for i in range(arr.shape[0]):
        for j in range(arr.shape[1]):
            res += arr[i, j]

    return res

##################考虑用太极实现



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-


import os
import sys
module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path)

import numpy as np
import pandas as pd
import pytest
from sklearn.linear_model import LogisticRegression

from autobmt.scorecard import ScoreCard, FeatureBin, WoeTransformer

np.random.seed(1)

# Create a testing dataframe and a scorecard model.

ab = np.array(list('ABCDEFG'))
feature = np.random.randint(10, size=500)
target = np.random.randint(2, size=500)
str_feat = ab[np.random.choice(7, 500)]

df = pd.DataFrame({
    'A': feature,
    'B': str_feat,
    'C': ab[np.random.choice(2, 500)],
    'D': np.ones(500),
})

df.to_csv('/Users/ryanzheng/PycharmProjects/autoscorecard/tests/genrate_data.csv', index=False)

card_config = {
    'A': {
        '[-inf ~ 3)': 100,
        '[3 ~ 5)': 200,
        '[5 ~ 8)': 300,
        '[8 ~ inf)': 400,
        'nan': 500,
    },
    'B': {
        ','.join(list('ABCD')): 200,
        ','.join(list('EF')): 400,
        'else': 500,
    },
    'C': {
        'A': 200,
        'B': 100,
    },
}

# card_config = {
# 	'A': {
# 		'[-inf ~ 2)': [142.27, 0.053086897685008266, 0.8196986054553417],
# 		'[2 ~ 5)': [151.36, -0.0750378915742799, 0.8196986054553417],
# 		'[5 ~ 6)': [118.09, 0.3938235330926786, 0.8196986054553417],
# 		'[6 ~ 7)': [178.21, -0.453474327294525, 0.8196986054553417],
# 		'[7 ~ inf)': [140.09, 0.08376005844476284, 0.8196986054553417]
# 	},
# 	'B': {
# 		'G,A': [162.08, -0.21634453400557532, 0.8567329286858805],
# 		'D': [159.24, -0.17806234743455837, 0.8567329286858805],
# 		'C': [152.82, -0.09149433112609942, 0.8567329286858805],
# 		'F': [134.04, 0.16171131179570847, 0.8567329286858805],
# 		'B,E': [123.68, 0.30136642227076027, 0.8567329286858805]
# 	},
# 	'C': {
# 		'A': [148.83, -0.07854594304644223, 0.4109435633463568],
# 		'B': [142.95, 0.08664766595989656, 0.4109435633463568]
# 	}
# }


combiner = FeatureBin()
bins = combiner.fit_transform(df, target, n_bins=5, is_need_monotonic=False)
woe_transer = WoeTransformer()
woe = woe_transer.fit_transform(bins, target)

lr = LogisticRegression()
lr.fit(woe, target)

# create a score card
card = ScoreCard(
    combiner=combiner,
    transer=woe_transer,
)

card.fit(woe, target)

FUZZ_THRESHOLD = 1e-6
TEST_SCORE = pytest.approx(414.0772336534059, FUZZ_THRESHOLD)


def test_representation():
    repr(card)


def test_load():
    card = ScoreCard().load(card_config)
    score = card.predict(df)
    assert score[200] == 600


def test_load_after_init_combiner():
    card = ScoreCard(
        combiner=combiner,
        transer=woe_transer,
    )
    card.load(card_config)
    score = card.predict(df)
    assert score[200] == 600


def test_proba_to_score():
    model = LogisticRegression()
    model.fit(woe, target)

    proba = model.predict_proba(woe)[:, 1]
    score = card.proba_to_score(proba)
    assert score[404] == TEST_SCORE


def test_score_to_prob():
    score = card.predict(df)
    proba = card.score_to_proba(score)
    assert proba[404] == pytest.approx(0.4673929989138551, FUZZ_THRESHOLD)


def test_predict():
    score = card.predict(df)
    assert score[404] == TEST_SCORE


def test_predict_proba():
    proba = card.predict_proba(df)
    assert proba[404, 1] == pytest.approx(0.4673929989138551, FUZZ_THRESHOLD)


def test_predict_sub_score():
    score, sub = card.predict(df, return_sub=True)
    assert sub.loc[250, 'B'] == pytest.approx(147.10499540339703, FUZZ_THRESHOLD)


def test_woe_to_score():
    score = card.woe_to_score(woe)
    score = np.sum(score, axis=1)
    assert score[404] == TEST_SCORE


def test_woe_sum_to_score():
    score = card.woe_sum_to_score(woe)
    assert score[404] == TEST_SCORE


def test_bin_to_score():
    score = card.bin_to_score(bins)
    assert score[404] == TEST_SCORE


def test_export_map():
    card_map = card.export()
    assert card_map['B']['D'][0] == 144.74


def test_card_map():
    config = card.export()
    card_from_map = ScoreCard().load(config)
    score = card_from_map.predict(df)
    assert score[404] == 414.0799999999999


def test_card_map_with_else():
    card_from_map = ScoreCard().load(card_config)
    score = card_from_map.predict(df)
    assert score[80] == 1000


def test_export_frame():
    card = ScoreCard().load(card_config)
    frame = card.export(to_dataframe=True)
    rows = frame[(frame['feature'] == 'B') & (frame['value'] == 'else')].reset_index()
    assert rows.loc[0, 'score'] == 500


def test_card_combiner_number_not_match():
    c = combiner.export(bin_format=False)
    c['A'] = [0, 3, 6, 8]
    com = FeatureBin().manual_bin(c)
    bins = com.transform(df)
    woe_transer = WoeTransformer()
    woe = woe_transer.fit_transform(bins, target)

    card = ScoreCard(
        combiner=com,
        transer=woe_transer,
    )

    with pytest.raises(Exception) as e:
        # will raise an exception when fitting a card
        card.fit(woe, target)

    assert '\'A\' is not matched' in str(e.value)


def test_card_combiner_str_not_match():
    c = combiner.export(bin_format=False)
    c['C'] = [['A'], ['B'], ['C']]
    com = FeatureBin().manual_bin(c)
    bins = com.transform(df)
    woe_transer = WoeTransformer()
    woe = woe_transer.fit_transform(bins, target)

    card = ScoreCard(
        combiner=com,
        transer=woe_transer,
    )

    with pytest.raises(Exception) as e:
        # will raise an exception when fitting a card
        card.fit(woe, target)

    assert '\'C\' is not matched' in str(e.value)


def test_card_with_less_X():
    x = woe.drop(columns='A')
    card = ScoreCard(
        combiner=combiner,
        transer=woe_transer,
    )

    card.fit(x, target)
    assert card.predict(df)[200] == pytest.approx(417.43225487849736, FUZZ_THRESHOLD)


def test_card_predict_with_unknown_feature():
    np.random.seed(9)
    unknown_df = df.copy()
    unknown_df.loc[200, 'C'] = 'U'
    assert card.predict(unknown_df)[200] == pytest.approx(416.4430811052408, FUZZ_THRESHOLD)


def test_card_predict_with_unknown_feature_default_max():
    np.random.seed(9)
    unknown_df = df.copy()
    unknown_df.loc[200, 'C'] = 'U'
    score, sub = card.predict(unknown_df, default='max', return_sub=True)

    assert sub.loc[200, 'C'] == card['C']['scores'].max()
    assert score[200] == pytest.approx(421.3399668050622, FUZZ_THRESHOLD)


def test_card_predict_with_unknown_feature_default_with_value():
    np.random.seed(9)
    unknown_df = df.copy()
    unknown_df.loc[200, 'C'] = 'U'
    score, sub = card.predict(unknown_df, default=42, return_sub=True)

    assert sub.loc[200, 'C'] == 42
    assert score[200] == pytest.approx(327.2767487322905, FUZZ_THRESHOLD)




#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: category_woe_encoder.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-09-21
'''

import json
import operator
import sys

import numpy as np
import pandas as pd
from joblib import Parallel, delayed


def category_2_woe(df, category_cols=[], target='target'):
    """
    每个类别都会转成woe值。缺失值不转，即还是为缺失值。在考虑到未来如果有新类别，给予other对应woe为0
    Args:
        df (DataFrame):
        category_cols (list): 类别变量list
        target:

    Returns:

    """
    var_value_woe = {}
    for i in category_cols:
        bin_g = df.groupby(by=i)[target].agg([('total_cnt', 'count'), ('bad_cnt', 'sum')])
        bin_g['good_cnt'] = bin_g['total_cnt'] - bin_g['bad_cnt']
        bad_count = sum(bin_g['bad_cnt'])
        good_count = sum(bin_g['good_cnt'])
        # bin_g['bad_rate'] = bin_g['bad_cnt'] / sum(bin_g['bad_cnt'])
        bin_g['bad_rate'] = bin_g['bad_cnt'].map(lambda x: 1 / bad_count if x == 0 else x / bad_count)
        # bin_g['good_rate'] = bin_g['good_cnt'] / sum(bin_g['good_cnt'])
        bin_g['good_rate'] = bin_g['good_cnt'].map(lambda x: 1 / good_count if x == 0 else x / good_count)
        # bin_g['good_rate'].replace({0: 0.0000000001}, inplace=True)  # good_rate为0的情况下，woe算出来是-inf。即将0使用一个极小数替换
        # bin_g['woe'] = bin_g.apply(lambda x: 0.0 if x['bad_rate'] == 0 else np.log(x['good_rate'] / x['bad_rate']),
        #                            axis=1)
        bin_g['woe'] = bin_g.apply(lambda x: np.log(x['good_rate'] / x['bad_rate']), axis=1)

        value_woe = bin_g['woe'].to_dict()
        value_woe['other'] = 0  # 未来有新类别的情况下，woe值给予0
        var_value_woe[i] = value_woe

    return var_value_woe


def bin_to_woe(df, var_bin_woe_dict):
    """
    根据传进来的var_bin_woe_dict对原始值进行映射。
    如在var_bin_woe_dict没有的类别（数据集中新出现的类别，归为到other这类）同时var_bin_woe_dict中得有other该类别对应的woe值
    如果var_bin_woe_dict中没有other该类别对应的woe值，即数据集中新出现的类别归为缺失值，即新出现的类别没有woe值
    Args:
        df:
        var_bin_woe_dict (dict):

    Returns:

    """

    for feature, bin_woe in var_bin_woe_dict.items():
        df[feature] = df[feature].map(
            lambda x: x if (x in bin_woe.keys() or x is np.nan or pd.isna(x)) else 'other')
        df[feature] = df[feature].map(bin_woe)

    return df


def data_to_bin(df, bins_dict={}):
    """
    原始数据根据bins_dict进行分箱
    Args:
        df:
        bins_dict: 分箱字典, 形如{'D157': [-999, 1.0, 2.0, 3.0, 5.0, inf]}

    Returns:

    """

    if not isinstance(bins_dict, dict):
        assert '请传入类似 {\'D157\': [-999, 1.0, 2.0, 3.0, 5.0, inf]}'

    data_with_bins = Parallel(n_jobs=-1)(
        delayed(pd.cut)(df[col], bins=bins, right=False, retbins=True) for col, bins in bins_dict.items())
    data_bin = pd.DataFrame([i[0].astype(str) for i in data_with_bins]).T
    b_dict = dict([(i[0].name, i[1].tolist()) for i in data_with_bins])
    if not operator.eq(bins_dict, b_dict):
        assert '传入的分箱和应用后的分箱不对等，请联系开发者'

    return data_bin


def transform(df, var_bin_woe_dict, bins_dict={}):
    """

    Args:
        df:
        var_bin_woe_dict (dict): 形如{"Sex": {"female": -1.5298770033401874, "male": 0.9838327092415774}, "Embarked": {"C": -0.694264203516269, "S": 0.1977338357888416, "other": -0.030202603851420356}}
        bins_dict:

    Returns:
        df (DataFrame): 转换woe后的数据集
    """

    df_ = df.copy()
    if bins_dict:
        df_ = data_to_bin(df, bins_dict=bins_dict)
    return bin_to_woe(df_, var_bin_woe_dict)


def category_2_woe_save(var_value_woe, path=None):
    if path is None:
        path = sys.path[0]

    with open(path + 'category_var_value_woe.json', 'w') as f:
        json.dump(var_value_woe, f)


def category_2_woe_load(path=None):
    with open(path + 'category_var_value_woe.json', 'r') as f:
        var_value_woe = json.load(f)
    return var_value_woe


#########################################测试代码

if __name__ == "__main__":
    from autobmt.utils import select_features_dtypes

    #####读取数据
    to_model_data_path = '../tests/tutorial_data.csv'
    cust_id = 'APP_ID_C'
    target = 'target'  # 目标变量
    data_type = 'type'
    all_data = pd.read_csv(to_model_data_path)
    train_data = all_data[all_data['type'] == 'train']

    n_cols, c_cols, d_cols = select_features_dtypes(train_data, exclude=[cust_id, target, data_type])
    print('数值特征个数: {}'.format(len(n_cols)))
    print('字符特征个数: {}'.format(len(c_cols)))
    print('日期特征个数: {}'.format(len(d_cols)))

    category_2_woe_save_path = "../tests/"

    print("类别变量数据处理前", all_data[c_cols])
    if c_cols:
        print('类别变量数据处理')
        # train_data.loc[:, category_cols] = train_data.loc[:, category_cols].fillna('miss')
        # test_data.loc[:, category_cols] = test_data.loc[:, category_cols].fillna('miss')

        var_value_woe = category_2_woe(train_data, c_cols, target=target)
        category_2_woe_save(var_value_woe, '{}'.format(category_2_woe_save_path))
        # var_value_woe = category_2_woe_load('{}'.format(output_dir))
        train_data = transform(train_data, var_value_woe)
        all_data = transform(all_data, var_value_woe)

    print("类别变量数据处理后", all_data[c_cols])
